import scipy
import numpy as np
import pytest
import qutip


def is_eigen_set(oper, vals, vecs):
    for val, vec in zip(vals, vecs):
        assert abs(vec.norm() - 1) < 1e-13
        assert abs(qutip.expect(oper, vec) - val) < 1e-13


@pytest.mark.parametrize(["sparse", 'dtype'], [
    pytest.param(True, 'csr', id="sparse"),
    pytest.param(False, 'csr', id="sparse2dense"),
    pytest.param(False, 'dense', id="dense"),
])
def test_eigen_known_oper(sparse, dtype):
    N = qutip.num(10, dtype=dtype)
    spvals, spvecs = N.eigenstates(sparse=sparse)
    expected = np.arange(10)
    is_eigen_set(N, spvals, spvecs)
    np.testing.assert_allclose(spvals, expected, atol=1e-13)


@pytest.mark.parametrize(["sparse", 'dtype'], [
    pytest.param(True, 'csr', id="sparse"),
    pytest.param(False, 'csr', id="sparse2dense"),
    pytest.param(False, 'dense', id="dense"),
])
@pytest.mark.parametrize(["rand"], [
    pytest.param(qutip.rand_herm, id="hermitian"),
    pytest.param(qutip.rand_unitary, id="non-hermitian"),
])
@pytest.mark.parametrize("order", ['low', 'high'])
def test_eigen_rand_oper(rand, sparse, dtype, order):
    H = rand(10, dtype=dtype)
    spvals, spvecs = H.eigenstates(sparse=sparse, sort=order)
    sp_energies = H.eigenenergies(sparse=sparse, sort=order)
    if order == 'low':
        assert np.all(np.diff(spvals).real >= 0)
    else:
        assert np.all(np.diff(spvals).real <= 0)
    is_eigen_set(H, spvals, spvecs)
    np.testing.assert_allclose(spvals, sp_energies, atol=5e-15)


@pytest.mark.parametrize(["sparse", 'dtype'], [
    pytest.param(True, 'csr', id="sparse"),
    pytest.param(False, 'csr', id="sparse2dense"),
    pytest.param(False, 'dense', id="dense"),
])
@pytest.mark.parametrize("rand", [
    pytest.param(qutip.rand_herm, id="hermitian"),
    pytest.param(qutip.rand_unitary, id="non-hermitian"),
])
@pytest.mark.parametrize("order", ['low', 'high'])
@pytest.mark.parametrize("N", [1, 5, 8, 9])
def test_FewState(rand, sparse, dtype, order, N):
    H = rand(10, dtype=dtype)
    all_spvals = H.eigenenergies(sparse=sparse, sort=order)
    spvals, spvecs = H.eigenstates(sparse=sparse, sort=order, eigvals=N)
    assert np.allclose(all_spvals[:N], spvals)
    is_eigen_set(H, spvals, spvecs)
    if order == 'low':
        assert np.all(np.diff(spvals).real >= 0)
    else:
        assert np.all(np.diff(spvals).real <= 0)


@pytest.mark.parametrize(["sparse", 'dtype'], [
    pytest.param(True, 'csr', id="sparse"),
    pytest.param(False, 'csr', id="sparse2dense"),
    pytest.param(False, 'dense', id="dense"),
])
@pytest.mark.parametrize(["rand"], [
    pytest.param(qutip.rand_herm, id="hermitian"),
    pytest.param(qutip.rand_unitary, id="non-hermitian"),
])
@pytest.mark.parametrize("order", ['low', 'high'])
@pytest.mark.parametrize("N", [1, 5, 8, 9])
def test_ValsOnly(rand, sparse, dtype, order, N):
    H = rand(10, dtype=dtype)
    all_spvals = H.eigenenergies(sparse=sparse, sort=order)
    spvals = H.eigenenergies(sparse=sparse, sort=order, eigvals=N)
    assert np.allclose(all_spvals[:N], spvals)
    if order == 'low':
        assert np.all(np.diff(spvals).real >= 0)
    else:
        assert np.all(np.diff(spvals).real <= 0)


@pytest.mark.parametrize(["sparse", 'dtype'], [
    pytest.param(True, 'csr', id="sparse"),
    pytest.param(False, 'csr', id="sparse2dense"),
    pytest.param(False, 'dense', id="dense"),
])
def test_eigen_small(sparse, dtype):
    H = (qutip.sigmax() + qutip.sigmaz()).to(dtype)
    all_spvals = H.eigenenergies(sparse=sparse)
    spvals, spvecs = H.eigenstates(sparse=sparse, eigvals=1)
    assert np.abs(all_spvals[0] - spvals[0]) <= 1e-14
    is_eigen_set(H, spvals, spvecs)


def test_BigDenseValsOnly():
    """
    This checks eigenvalue calculation for large dense matrices, which
    historically have had instabilities with certain OS and BLAS combinations
    (see e.g. #1288 and #1495).
    """
    dimension = 2000
    # Allow an average absolute tolerance for each eigenvalue; we expect
    # uncertainty in the sum to add in quadrature.
    tol = 1e-12 * np.sqrt(dimension)
    H = qutip.rand_herm(dimension, density=1e-2)
    spvals = H.eigenenergies()
    assert abs(H.tr() - spvals.sum()) < tol



__all__ = [
    'correlation_3op',
    'correlation_2op_1t', 'correlation_2op_2t', 'correlation_3op_1t',
    'correlation_3op_2t', 'coherence_function_g1', 'coherence_function_g2'
]

import numpy as np
import scipy.fftpack

from ..core import (
    qeye, Qobj, QobjEvo, liouvillian, spre, unstack_columns, stack_columns,
    tensor, qzero, expect
)
from .mesolve import MESolver
from .mcsolve import MCSolver
from .brmesolve import BRSolver
from .heom.bofin_solvers import HEOMSolver

from .steadystate import steadystate
from ..ui.progressbar import progess_bars

# -----------------------------------------------------------------------------
# PUBLIC API
# -----------------------------------------------------------------------------

# low level correlation


def correlation_2op_1t(H, state0, taulist, c_ops, a_op, b_op,
                       solver="me", reverse=False, args={},
                       options={}):
    r"""
    Calculate the two-operator one-time correlation function:
    :math:`\left<A(\tau)B(0)\right>`
    along one time axis using the quantum regression theorem and the evolution
    solver indicated by the `solver` parameter.

    Parameters
    ----------

    H : :class:`Qobj`, :class:`QobjEvo`
        System Hamiltonian, may be time-dependent for solver choice of `me`.
    state0 : :class:`Qobj`
        Initial state density matrix :math:`\rho(t_0)` or state vector
        :math:`\psi(t_0)`. If 'state0' is 'None', then the steady state will
        be used as the initial state. The 'steady-state' is only implemented
        if ``c_ops`` are provided and the Hamiltonian is constant.
    taulist : array_like
        List of times for :math:`\tau`. taulist must be positive and contain
        the element `0`.
    c_ops : list of {:class:`Qobj`, :class:`QobjEvo`}
        List of collapse operators
    a_op : :class:`Qobj`, :class:`QobjEvo`
        Operator A.
    b_op : :class:`Qobj`, :class:`QobjEvo`
        Operator B.
    reverse : bool {False}
        If `True`, calculate :math:`\left<A(t)B(t+\tau)\right>` instead of
        :math:`\left<A(t+\tau)B(t)\right>`.
    solver : str {'me', 'es'}
        Choice of solver, `me` for master-equation, and `es` for exponential
        series. `es` is equivalent to `me` with ``options={"method": "diag"}``.
    options : dict, optional
        Options for the solver. Only used with `me` solver.

    Returns
    -------
    corr_vec : ndarray
        An array of correlation values for the times specified by `taulist`.

    See Also
    --------
    :func:`correlation_3op` :
        Similar function supporting various solver types.

    References
    ----------
    See, Gardiner, Quantum Noise, Section 5.2.

    """
    solver = _make_solver(H, c_ops, args, options, solver)

    if reverse:
        A_op, B_op, C_op = a_op, b_op, 1
    else:
        A_op, B_op, C_op = 1, a_op, b_op
    if state0 is None:
        state0 = steadystate(H, c_ops)

    return correlation_3op(solver, state0, [0], taulist, A_op, B_op, C_op)[0]


def correlation_2op_2t(H, state0, tlist, taulist, c_ops, a_op, b_op,
                       solver="me", reverse=False, args={},
                       options={}):
    r"""
    Calculate the two-operator two-time correlation function:
    :math:`\left<A(t+\tau)B(t)\right>`
    along two time axes using the quantum regression theorem and the
    evolution solver indicated by the ``solver`` parameter.

    Parameters
    ----------
    H : :class:`Qobj`, :class:`QobjEvo`
        System Hamiltonian, may be time-dependent for solver choice of `me`.
    state0 : :class:`Qobj`
        Initial state density matrix :math:`\rho(t_0)` or state vector
        :math:`\psi(t_0)`. If 'state0' is 'None', then the steady state will
        be used as the initial state. The 'steady-state' is only implemented
        if ``c_ops`` are provided and the Hamiltonian is constant.
    tlist : array_like
        List of times for :math:`t`. tlist must be positive and contain the
        element `0`. When taking steady-steady correlations only one ``tlist``
        value is necessary, i.e. when :math:`t \rightarrow \infty`.
        If ``tlist`` is ``None``, ``tlist=[0]`` is assumed.
    taulist : array_like
        List of times for :math:`\tau`. taulist must be positive and contain
        the element `0`.
    c_ops : list of {:class:`Qobj`, :class:`QobjEvo`}
        List of collapse operators
    a_op : :class:`Qobj`, :class:`QobjEvo`
        Operator A.
    b_op : :class:`Qobj`, :class:`QobjEvo`
        Operator B.
    reverse : bool {False}
        If `True`, calculate :math:`\left<A(t)B(t+\tau)\right>` instead of
        :math:`\left<A(t+\tau)B(t)\right>`.
    solver : str {'me', 'es'}
        Choice of solver, `me` for master-equation, and `es` for exponential
        series. `es` is equivalent to `me` with ``options={"method": "diag"}``.
    options : dict, optional
        Options for the solver. Only used with `me` solver.

    Returns
    -------
    corr_mat : ndarray
        An 2-dimensional array (matrix) of correlation values for the times
        specified by `tlist` (first index) and `taulist` (second index).

    See Also
    --------
    :func:`correlation_3op` :
        Similar function supporting various solver types.

    References
    ----------
    See, Gardiner, Quantum Noise, Section 5.2.

    """
    solver = _make_solver(H, c_ops, args, options, solver)
    if tlist is None:
        tlist = [0]
    if state0 is None:
        state0 = steadystate(H, c_ops)

    if reverse:
        A_op, B_op, C_op = a_op, b_op, 1
    else:
        A_op, B_op, C_op = 1, a_op, b_op

    return correlation_3op(solver, state0, tlist, taulist, A_op, B_op, C_op)


def correlation_3op_1t(H, state0, taulist, c_ops, a_op, b_op, c_op,
                       solver="me", args={},
                       options={}):
    r"""
    Calculate the three-operator two-time correlation function:
    :math:`\left<A(0)B(\tau)C(0)\right>` along one time axis using the
    quantum regression theorem and the evolution solver indicated by the
    `solver` parameter.

    Note: it is not possibly to calculate a physically meaningful correlation
    of this form where :math:`\tau<0`.

    Parameters
    ----------
    H : :class:`Qobj`, :class:`QobjEvo`
        System Hamiltonian, may be time-dependent for solver choice of `me`.
    state0 : :class:`Qobj`
        Initial state density matrix :math:`\rho(t_0)` or state vector
        :math:`\psi(t_0)`. If 'state0' is 'None', then the steady state will
        be used as the initial state. The 'steady-state' is only implemented
        if ``c_ops`` are provided and the Hamiltonian is constant.
    taulist : array_like
        List of times for :math:`\tau`. taulist must be positive and contain
        the element `0`.
    c_ops : list of {:class:`Qobj`, :class:`QobjEvo`}
        List of collapse operators
    a_op : :class:`Qobj`, :class:`QobjEvo`
        Operator A.
    b_op : :class:`Qobj`, :class:`QobjEvo`
        Operator B.
    c_op : :class:`Qobj`, :class:`QobjEvo`
        Operator C.
    solver : str {'me', 'es'}
        Choice of solver, `me` for master-equation, and `es` for exponential
        series. `es` is equivalent to `me` with ``options={"method": "diag"}``.
    options : dict, optional
        Options for the solver. Only used with `me` solver.

    Returns
    -------
    corr_vec : array
        An array of correlation values for the times specified by `taulist`.

    See Also
    --------
    :func:`correlation_3op` :
        Similar function supporting various solver types.

    References
    ----------
    See, Gardiner, Quantum Noise, Section 5.2.

    """
    solver = _make_solver(H, c_ops, args, options, solver)
    if state0 is None:
        state0 = steadystate(H, c_ops)
    return correlation_3op(solver, state0, [0], taulist, a_op, b_op, c_op)[0]


def correlation_3op_2t(H, state0, tlist, taulist, c_ops, a_op, b_op, c_op,
                       solver="me", args={},
                       options={}):
    r"""
    Calculate the three-operator two-time correlation function:
    :math:`\left<A(t)B(t+\tau)C(t)\right>` along two time axes using the
    quantum regression theorem and the evolution solver indicated by the
    `solver` parameter.

    Note: it is not possibly to calculate a physically meaningful correlation
    of this form where :math:`\tau<0`.

    Parameters
    ----------
    H : :class:`Qobj`, :class:`QobjEvo`
        System Hamiltonian, may be time-dependent for solver choice of `me`.
    state0 : :class:`Qobj`
        Initial state density matrix :math:`\rho(t_0)` or state vector
        :math:`\psi(t_0)`. If 'state0' is 'None', then the steady state will
        be used as the initial state. The 'steady-state' is only implemented
        if ``c_ops`` are provided and the Hamiltonian is constant.
    tlist : array_like
        List of times for :math:`t`. tlist must be positive and contain the
        element `0`. When taking steady-steady correlations only one tlist
        value is necessary, i.e. when :math:`t \rightarrow \infty`.
        If ``tlist`` is ``None``, ``tlist=[0]`` is assumed.
    taulist : array_like
        List of times for :math:`\tau`. taulist must be positive and contain
        the element `0`.
    c_ops : list of {:class:`Qobj`, :class:`QobjEvo`}
        List of collapse operators
    a_op : :class:`Qobj`, :class:`QobjEvo`
        Operator A.
    b_op : :class:`Qobj`, :class:`QobjEvo`
        Operator B.
    c_op : :class:`Qobj`, :class:`QobjEvo`
        Operator C.
    solver : str {'me', 'es'}
        Choice of solver, `me` for master-equation, and `es` for exponential
        series. `es` is equivalent to `me` with ``options={"method": "diag"}``.
    options : dict, optional
        Options for the solver. Only used with `me` solver.

    Returns
    -------
    corr_mat : array
        An 2-dimensional array (matrix) of correlation values for the times
        specified by `tlist` (first index) and `taulist` (second index).

    See Also
    --------
    :func:`correlation_3op` :
        Similar function supporting various solver types.

    References
    ----------

    See, Gardiner, Quantum Noise, Section 5.2.

    """
    solver = _make_solver(H, c_ops, args, options, solver)

    if tlist is None:
        tlist = [0]
    if state0 is None:
        state0 = steadystate(H, c_ops)

    return correlation_3op(solver, state0, tlist, taulist, a_op, b_op, c_op)


# high level correlation

def coherence_function_g1(
    H, state0, taulist, c_ops, a_op, solver="me", args={}, options={}
):
    r"""
    Calculate the normalized first-order quantum coherence function:

    .. math::

        g^{(1)}(\tau) =
        \frac{\langle A^\dagger(\tau)A(0)\rangle}
        {\sqrt{\langle A^\dagger(\tau)A(\tau)\rangle
                \langle A^\dagger(0)A(0)\rangle}}

    using the quantum regression theorem and the evolution solver indicated by
    the `solver` parameter.

    Parameters
    ----------
    H : :class:`Qobj`, :class:`QobjEvo`
        System Hamiltonian, may be time-dependent for solver choice of `me`.
    state0 : :class:`Qobj`
        Initial state density matrix :math:`\rho(t_0)` or state vector
        :math:`\psi(t_0)`. If 'state0' is 'None', then the steady state will
        be used as the initial state. The 'steady-state' is only implemented
        if ``c_ops`` are provided and the Hamiltonian is constant.
    taulist : array_like
        List of times for :math:`\tau`. taulist must be positive and contain
        the element `0`.
    c_ops : list of {:class:`Qobj`, :class:`QobjEvo`}
        List of collapse operators
    a_op : :class:`Qobj`, :class:`QobjEvo`
        Operator A.
    solver : str {'me', 'es'}
        Choice of solver, `me` for master-equation, and `es` for exponential
        series. `es` is equivalent to `me` with ``options={"method": "diag"}``.
    options : dict, optional
        Options for the solver. Only used with `me` solver.

    Returns
    -------
    g1, G1 : tuple
        The normalized and unnormalized second-order coherence function.

    """
    solver = _make_solver(H, c_ops, args, options, solver)

    # first calculate the photon number
    if state0 is None:
        state0 = steadystate(H, c_ops)
        n = np.array([expect(state0, a_op.dag() * a_op)])
    else:
        n = solver.run(state0, taulist, e_ops=[a_op.dag() * a_op]).expect[0]

    # calculate the correlation function G1 and normalize with n to obtain g1
    G1 = correlation_3op(solver, state0, [0], taulist, None, a_op.dag(), a_op)[0]

    g1 = G1 / np.sqrt(n[0] * np.array(n))[0]
    return g1, G1


def coherence_function_g2(H, state0, taulist, c_ops, a_op, solver="me",
                          args={}, options={}):
    r"""
    Calculate the normalized second-order quantum coherence function:

    .. math::

         g^{(2)}(\tau) =
        \frac{\langle A^\dagger(0)A^\dagger(\tau)A(\tau)A(0)\rangle}
        {\langle A^\dagger(\tau)A(\tau)\rangle
         \langle A^\dagger(0)A(0)\rangle}

    using the quantum regression theorem and the evolution solver indicated by
    the `solver` parameter.

    Parameters
    ----------
    H : :class:`Qobj`, :class:`QobjEvo`
        System Hamiltonian, may be time-dependent for solver choice of `me`.
    state0 : :class:`Qobj`
        Initial state density matrix :math:`\rho(t_0)` or state vector
        :math:`\psi(t_0)`. If 'state0' is 'None', then the steady state will
        be used as the initial state. The 'steady-state' is only implemented
        if ``c_ops`` are provided and the Hamiltonian is constant.
    taulist : array_like
        List of times for :math:`\tau`. taulist must be positive and contain
        the element `0`.
    c_ops : list
        List of collapse operators, may be time-dependent for solver choice of
        `me`.
    a_op : :class:`Qobj`
        Operator A.
    args : dict
        Dictionary of arguments to be passed to solver.
    solver : str {'me', 'es'}
        Choice of solver, `me` for master-equation, and `es` for exponential
        series. `es` is equivalent to `me` with ``options={"method": "diag"}``.
    options : dict, optional
        Options for the solver. Only used with `me` solver.

    Returns
    -------
    g2, G2 : tuple
        The normalized and unnormalized second-order coherence function.

    """
    solver = _make_solver(H, c_ops, args, options, solver)

    # first calculate the photon number
    if state0 is None:
        state0 = steadystate(H, c_ops)
        n = np.array([expect(state0, a_op.dag() * a_op)])
    else:
        n = solver.run(state0, taulist, e_ops=[a_op.dag() * a_op]).expect[0]

    # calculate the correlation function G2 and normalize with n to obtain g2
    G2 = correlation_3op(solver, state0, [0], taulist,
                         a_op.dag(), a_op.dag() * a_op, a_op)[0]

    g2 = G2 / (n[0] * np.array(n))
    return g2, G2


def _make_solver(H, c_ops, args, options, solver):
    H = QobjEvo(H, args=args)
    c_ops = [QobjEvo(c_op, args=args) for c_op in c_ops]
    if solver == "me":
        solver_instance = MESolver(H, c_ops, options=options)
    elif solver == "es":
        options = {"method": "diag"}
        solver_instance = MESolver(H, c_ops, options=options)
    elif solver == "mc":
        raise ValueError("MC solver for correlation has been removed")
    return solver_instance


def correlation_3op(solver, state0, tlist, taulist, A=None, B=None, C=None):
    r"""
    Calculate the three-operator two-time correlation function:

        :math:`\left<A(t)B(t+\tau)C(t)\right>`.

    from a open system :class:`Solver`.

    Note: it is not possible to calculate a physically meaningful correlation
    where :math:`\tau<0`.

    Parameters
    ----------
    solver : :class:`MESolver`, :class:`BRSolver`
        Qutip solver for an open system.
    state0 : :class:`Qobj`
        Initial state density matrix :math:`\rho(t_0)` or state vector
        :math:`\psi(t_0)`.
    tlist : array_like
        List of times for :math:`t`. tlist must be positive and contain the
        element `0`.
    taulist : array_like
        List of times for :math:`\tau`. taulist must be positive and contain
        the element `0`.
    A, B, C: :class:`Qobj`, :class:`QobjEvo`, optional, default=None
        Operators ``A``, ``B``, ``C`` from the equation ``<A(t)B(t+\tau)C(t)>``
        in the Schrodinger picture. They do not need to be all provided. For
        exemple, if ``A`` is not provided, ``<B(t+\tau)C(t)>`` is computed.

    Returns
    -------
    corr_mat : array
        An 2-dimensional array (matrix) of correlation values for the times
        specified by `tlist` (first index) and `taulist` (second index). If
        `tlist` is `None`, then a 1-dimensional array of correlation values
        is returned instead.
    """
    taulist = np.asarray(taulist)

    dims = state0.dims[0]
    A = QobjEvo(qeye(dims) if A in [None, 1] else A)
    B = QobjEvo(qeye(dims) if B in [None, 1] else B)
    C = QobjEvo(qeye(dims) if C in [None, 1] else C)

    if isinstance(solver, (MESolver, BRSolver)):
        out = _correlation_3op_dm(solver, state0, tlist, taulist, A, B, C)
    elif isinstance(solver, MCSolver):
        raise TypeError("Monte Carlo support for correlation was removed. "
                        "Please, tell us on GitHub issues if you need it!")
    elif isinstance(solver, HEOMSolver):
        raise TypeError("HEOM is not supported by correlation. "
                        "Please, tell us on GitHub issues if you need it!")
    else:
        raise TypeError("Only solvers able to evolve density matrices"
                        " are supported.")

    return out


def _correlation_3op_dm(solver, state0, tlist, taulist, A, B, C):
    old_opt = solver.options.copy()
    try:
        # We don't want to modify the solver
        # TODO: Solver could have a ``with`` or ``copy``.
        solver.options["normalize_output"] = False
        solver.options["progress_bar"] = False

        progress_bar = progess_bars[old_opt['progress_bar']]()
        progress_bar.start(len(taulist) + 1, **old_opt['progress_kwargs'])
        rho_t = solver.run(state0, tlist).states
        corr_mat = np.zeros([np.size(tlist), np.size(taulist)], dtype=complex)
        progress_bar.update()

        for t_idx, rho in enumerate(rho_t):
            t = tlist[t_idx]
            corr_mat[t_idx, :] = solver.run(
                C(t) @ rho @ A(t),
                taulist + t,
                e_ops=B
            ).expect[0]
            progress_bar.update()
        progress_bar.finished()

    finally:
        solver.options = old_opt

    return corr_mat



import numpy as np
import time
import pytest
import threading

from qutip.solver.parallel import (
    parallel_map, serial_map, loky_pmap, MapExceptions
)


def _func1(x):
    return x**2


def _func2(x, a, b, c, d=0, e=0, f=0):
    assert d > 0
    assert e > 0
    assert f > 0
    time.sleep(np.random.rand() * 0.1)  # random delay
    return x**2

@pytest.mark.parametrize('map', [
    pytest.param(parallel_map, id='parallel_map'),
    pytest.param(loky_pmap, id='loky_pmap'),
    pytest.param(serial_map, id='serial_map'),
])
@pytest.mark.parametrize('num_cpus',
                         [1, 2],
                         ids=['1', '2'])
def test_map(map, num_cpus):
    if map is loky_pmap:
        loky = pytest.importorskip("loky")

    args = (1, 2, 3)
    kwargs = {'d': 4, 'e': 5, 'f': 6}
    map_kw = {
        'job_timeout': threading.TIMEOUT_MAX,
        'timeout': threading.TIMEOUT_MAX,
        'num_cpus': num_cpus,
    }

    x = np.arange(10)
    y1 = [_func1(xx) for xx in x]

    y2 = map(_func2, x, args, kwargs, map_kw=map_kw)
    assert ((np.array(y1) == np.array(y2)).all())


@pytest.mark.parametrize('map', [
    pytest.param(parallel_map, id='parallel_map'),
    pytest.param(loky_pmap, id='loky_pmap'),
    pytest.param(serial_map, id='serial_map'),
])
@pytest.mark.parametrize('num_cpus',
                         [1, 2],
                         ids=['1', '2'])
def test_map_accumulator(map, num_cpus):
    if map is loky_pmap:
        loky = pytest.importorskip("loky")
    args = (1, 2, 3)
    kwargs = {'d': 4, 'e': 5, 'f': 6}
    map_kw = {
        'job_timeout': threading.TIMEOUT_MAX,
        'timeout': threading.TIMEOUT_MAX,
        'num_cpus': num_cpus,
    }
    y2 = []

    x = np.arange(10)
    y1 = [_func1(xx) for xx in x]

    map(_func2, x, args, kwargs, reduce_func=y2.append, map_kw=map_kw)
    assert ((np.array(sorted(y1)) == np.array(sorted(y2))).all())


class CustomException(Exception):
    pass


def func(i):
    if i % 2 == 1:
        raise CustomException(f"Error in subprocess {i}")
    return i


@pytest.mark.parametrize('map', [
    pytest.param(parallel_map, id='parallel_map'),
    pytest.param(loky_pmap, id='loky_pmap'),
    pytest.param(serial_map, id='serial_map'),
])
def test_map_pass_error(map):
    if map is loky_pmap:
        loky = pytest.importorskip("loky")

    with pytest.raises(CustomException) as err:
        map(func, range(10))
    assert "Error in subprocess" in str(err.value)


@pytest.mark.parametrize('map', [
    pytest.param(parallel_map, id='parallel_map'),
    pytest.param(loky_pmap, id='loky_pmap'),
    pytest.param(serial_map, id='serial_map'),
])
def test_map_store_error(map):
    if map is loky_pmap:
        loky = pytest.importorskip("loky")

    with pytest.raises(MapExceptions) as err:
        map(func, range(10), map_kw={"fail_fast": False})
    map_error = err.value
    assert "iterations failed" in str(map_error)
    for iter, error in map_error.errors.items():
        assert isinstance(error, CustomException)
        assert f"Error in subprocess {iter}" == str(error)
    for n, result in enumerate(map_error.results):
        if n % 2 == 0:
            # Passed
            assert result == n
        else:
            assert result is None



#cython: language_level=3
#cython: boundscheck=False, wraparound=False, initializedcheck=False

from libc.string cimport memset, memcpy

from libcpp cimport bool
from libcpp.algorithm cimport sort

cimport cython

from cpython cimport mem

import numbers
import warnings

import numpy as np
cimport numpy as cnp
import scipy.sparse
from scipy.sparse import csr_matrix as scipy_csr_matrix
try:
    from scipy.sparse.data import _data_matrix as scipy_data_matrix
except ImportError:
    # The file data was renamed to _data from scipy 1.8.0
    from scipy.sparse._data import _data_matrix as scipy_data_matrix
from scipy.linalg cimport cython_blas as blas

from qutip.core.data cimport base, Dense
from qutip.core.data.add cimport add_csr, sub_csr
from qutip.core.data.adjoint cimport adjoint_csr, transpose_csr, conj_csr
from qutip.core.data.mul cimport mul_csr, neg_csr
from qutip.core.data.matmul cimport matmul_csr
from qutip.core.data.trace cimport trace_csr
from qutip.core.data.tidyup cimport tidyup_csr
from .base import idxint_dtype
from qutip.settings import settings

cnp.import_array()

cdef extern from *:
    void PyArray_ENABLEFLAGS(cnp.ndarray arr, int flags)
    void *PyDataMem_NEW(size_t size)
    void PyDataMem_FREE(void *ptr)

# Very little should be exported on star-import, because most stuff will have
# naming collisions with other type modules.
__all__ = ['CSR']

cdef int _ONE = 1

cdef object _csr_matrix(data, indices, indptr, shape):
    """
    Factory method of scipy csr_matrix: we skip all the index type-checking
    because this takes tens of microseconds, and we already know we're in
    a sensible format.
    """
    cdef object out = scipy_csr_matrix.__new__(scipy_csr_matrix)
    # `_data_matrix` is the first object in the inheritance chain which
    # doesn't have a really slow __init__.
    scipy_data_matrix.__init__(out)
    out.data = data
    out.indices = indices
    out.indptr = indptr
    out._shape = shape
    return out


cdef class CSR(base.Data):
    """
    Data type for quantum objects storing its data in compressed sparse row
    (CSR) format.  This is similar to the `scipy` type
    `scipy.sparse.csr_matrix`, but significantly faster on many operations.
    You can retrieve a `scipy.sparse.csr_matrix` which views onto the same data
    using the `as_scipy()` method.
    """
    def __cinit__(self, *args, **kwargs):
        # By default, we want CSR to deallocate its memory (we depend on Cython
        # to ensure we don't deallocate a NULL pointer), and we only flip this
        # when we create a scipy backing.  Since creating the scipy backing
        # depends on knowing the shape, which happens _after_ data
        # initialisation and may throw an exception, it is better to have a
        # single flag that is set as soon as the pointers are assigned.
        self._deallocate = True

    def __init__(self, arg=None, shape=None, bint copy=True, bint tidyup=False):
        # This is the Python __init__ method, so we do not care that it is not
        # super-fast C access.  Typically Cython code will not call this, but
        # will use a factory method in this module or at worst, call
        # CSR.__new__ and manually set everything.  We must be very careful in
        # this function that the deallocation is set up correctly if any
        # exceptions occur.
        cdef size_t ptr
        cdef base.idxint col
        cdef object data, col_index, row_index
        if isinstance(arg, scipy.sparse.spmatrix):
            arg = arg.tocsr()
            if shape is not None and shape != arg.shape:
                raise ValueError("".join([
                    "shapes do not match: ", str(shape), " and ", str(arg.shape),
                ]))
            shape = arg.shape
            arg = (arg.data, arg.indices, arg.indptr)
        if not isinstance(arg, tuple):
            raise TypeError("arg must be a scipy matrix or tuple")
        if len(arg) != 3:
            raise ValueError("arg must be a (data, col_index, row_index) tuple")
        data = np.array(arg[0], dtype=np.complex128, copy=copy, order='C')
        col_index = np.array(arg[1], dtype=idxint_dtype, copy=copy, order='C')
        row_index = np.array(arg[2], dtype=idxint_dtype, copy=copy, order='C')
        # This flag must be set at the same time as data, col_index and
        # row_index are assigned.  These assignments cannot raise an exception
        # in user code due to the above three lines, but further code may.
        self._deallocate = False
        self.data = <double complex *> cnp.PyArray_GETPTR1(data, 0)
        self.col_index = <base.idxint *> cnp.PyArray_GETPTR1(col_index, 0)
        self.row_index = <base.idxint *> cnp.PyArray_GETPTR1(row_index, 0)
        self.size = (cnp.PyArray_SIZE(data)
                     if cnp.PyArray_SIZE(data) < cnp.PyArray_SIZE(col_index)
                     else cnp.PyArray_SIZE(col_index))
        if shape is None:
            warnings.warn("instantiating CSR matrix of unknown shape")
            # row_index contains an extra element which is nnz.  We assume the
            # smallest matrix which can hold all these values by iterating
            # through the columns.  This is slow and probably inaccurate, since
            # there could be columns containing zero (hence the warning).
            self.shape[0] = cnp.PyArray_DIMS(row_index)[0] - 1
            col = 1
            for ptr in range(self.size):
                col = self.col_index[ptr] if self.col_index[ptr] > col else col
            self.shape[1] = col
        else:
            if not isinstance(shape, tuple):
                raise TypeError("shape must be a 2-tuple of positive ints")
            if not (len(shape) == 2
                    and isinstance(shape[0], int)
                    and isinstance(shape[1], int)
                    and shape[0] > 0
                    and shape[1] > 0):
                raise ValueError("shape must be a 2-tuple of positive ints")
            self.shape = shape
        # Store a reference to the backing scipy matrix so it doesn't get
        # deallocated before us.
        self._scipy = _csr_matrix(data, col_index, row_index, self.shape)
        if tidyup:
            tidyup_csr(self, settings.core['auto_tidyup_atol'], True)

    def __reduce__(self):
        return (fast_from_scipy, (self.as_scipy(),))

    cpdef CSR copy(self):
        """
        Return a complete (deep) copy of this object.

        If the type currently has a scipy backing, such as that produced by
        `as_scipy`, this will not be copied.  The backing is a view onto our
        data, and a straight copy of this view would be incorrect.  We do not
        create a new view at copy time, since the user may only access this
        through a creation method, and creating it ahead of time would incur an
        unnecessary speed penalty for users who do not need it (including
        low-level C code).
        """
        cdef base.idxint nnz_ = nnz(self)
        cdef CSR out = empty_like(self)
        memcpy(out.data, self.data, nnz_*sizeof(double complex))
        memcpy(out.col_index, self.col_index, nnz_*sizeof(base.idxint))
        memcpy(out.row_index, self.row_index,
               (self.shape[0] + 1)*sizeof(base.idxint))
        return out

    cpdef object to_array(self):
        """
        Get a copy of this data as a full 2D, C-contiguous NumPy array.  This
        is not a view onto the data, and changes to new array will not affect
        the original data structure.
        """
        cdef cnp.npy_intp *dims = [self.shape[0], self.shape[1]]
        cdef object out = cnp.PyArray_ZEROS(2, dims, cnp.NPY_COMPLEX128, 0)
        cdef double complex [:, ::1] buffer = out
        cdef size_t row, ptr
        for row in range(self.shape[0]):
            for ptr in range(self.row_index[row], self.row_index[row + 1]):
                buffer[row, self.col_index[ptr]] = self.data[ptr]
        return out

    cpdef object as_scipy(self, bint full=False):
        """
        Get a view onto this object as a `scipy.sparse.csr_matrix`.  The
        underlying data structures are exposed, such that modifications to the
        `data`, `indices` and `indptr` buffers in the resulting object will
        modify this object too.

        If `full` is False (the default), the output array is squeezed so that
        the SciPy array sees only the filled elements.  If `full` is True,
        SciPy will see the full underlying buffers, which may include
        uninitialised elements.  Setting `full=True` is intended for
        Python-space factory methods.  In all other use cases, `full=False` is
        much less error prone.
        """
        # We store a reference to the scipy matrix not only for caching this
        # relatively expensive method, but also because we transferred
        # ownership of our data to the numpy arrays, and we can't allow them to
        # be collected while we're alive.
        if self._scipy is not None:
            return self._scipy
        cdef cnp.npy_intp length = self.size if full else nnz(self)
        data = cnp.PyArray_SimpleNewFromData(1, [length],
                                             cnp.NPY_COMPLEX128,
                                             self.data)
        indices = cnp.PyArray_SimpleNewFromData(1, [length],
                                                base.idxint_DTYPE,
                                                self.col_index)
        indptr = cnp.PyArray_SimpleNewFromData(1, [self.shape[0] + 1],
                                               base.idxint_DTYPE,
                                               self.row_index)
        PyArray_ENABLEFLAGS(data, cnp.NPY_ARRAY_OWNDATA)
        PyArray_ENABLEFLAGS(indices, cnp.NPY_ARRAY_OWNDATA)
        PyArray_ENABLEFLAGS(indptr, cnp.NPY_ARRAY_OWNDATA)
        self._deallocate = False
        self._scipy = _csr_matrix(data, indices, indptr, self.shape)
        return self._scipy

    cpdef CSR sort_indices(self):
        cdef Sorter sort
        cdef base.idxint ptr
        cdef size_t row, diff, size=0
        for row in range(self.shape[0]):
            diff = self.row_index[row + 1] - self.row_index[row]
            size = diff if diff > size else size
        sort = Sorter(size)
        for row in range(self.shape[0]):
            ptr = self.row_index[row]
            diff = self.row_index[row + 1] - ptr
            sort.inplace(self, ptr, diff)
        return self

    cpdef double complex trace(self):
        return trace_csr(self)

    cpdef CSR adjoint(self):
        return adjoint_csr(self)

    cpdef CSR conj(self):
        return conj_csr(self)

    cpdef CSR transpose(self):
        return transpose_csr(self)

    def __repr__(self):
        return "".join([
            "CSR(shape=", str(self.shape), ", nnz=", str(nnz(self)), ")",
        ])

    def __str__(self):
        return self.__repr__()

    def __dealloc__(self):
        # If we have a reference to a scipy type, then we've passed ownership
        # of the data to numpy, so we let it handle refcounting and we don't
        # need to deallocate anything ourselves.
        if not self._deallocate:
            return
        if self.data != NULL:
            PyDataMem_FREE(self.data)
        if self.col_index != NULL:
            PyDataMem_FREE(self.col_index)
        if self.row_index != NULL:
            PyDataMem_FREE(self.row_index)


cpdef CSR fast_from_scipy(object sci):
    """
    Fast path construction from scipy.sparse.csr_matrix.  This does _no_ type
    checking on any of the inputs, and should consequently be considered very
    unsafe.  This is primarily for use in the unpickling operation.
    """
    cdef CSR out = CSR.__new__(CSR)
    out.shape = sci.shape
    out._deallocate = False
    out._scipy = sci
    out.data = <double complex *> cnp.PyArray_GETPTR1(sci.data, 0)
    out.col_index = <base.idxint *> cnp.PyArray_GETPTR1(sci.indices, 0)
    out.row_index = <base.idxint *> cnp.PyArray_GETPTR1(sci.indptr, 0)
    out.size = cnp.PyArray_SIZE(sci.data)
    return out


cpdef CSR copy_structure(CSR matrix):
    """
    Return a copy of the input matrix with identical `col_index` and
    `row_index` matrices, and an allocated, but empty, `data`.  The returned
    pointers are all separately allocated, but contain the same information.

    This is intended for unary functions on CSR types that maintain the exact
    structure, but modify each non-zero data element without change their
    location.
    """
    cdef CSR out = empty_like(matrix)
    memcpy(out.col_index, matrix.col_index, nnz(matrix) * sizeof(base.idxint))
    memcpy(out.row_index, matrix.row_index, (matrix.shape[0] + 1)*sizeof(base.idxint))
    return out


cpdef inline base.idxint nnz(CSR matrix) nogil:
    """Get the number of non-zero elements of a CSR matrix."""
    return matrix.row_index[matrix.shape[0]]


cdef bool _sorter_cmp_ptr(base.idxint *i, base.idxint *j) nogil:
    return i[0] < j[0]

cdef bool _sorter_cmp_struct(_data_col x, _data_col y) nogil:
    return x.col < y.col

ctypedef fused _swap_data:
    double complex
    base.idxint

cdef inline void _sorter_swap(_swap_data *a, _swap_data *b) nogil:
    a[0], b[0] = b[0], a[0]

cdef class Sorter:
    # Look on my works, ye mighty, and despair!
    #
    # This class has hard-coded sorts for up to three elements, for both
    # copying and in-place varieties.  Everything above that we delegate to a
    # proper sorting algorithm.
    def __init__(self, size_t size):
        self.size = size

    cdef void inplace(self, CSR matrix, base.idxint ptr, size_t size) nogil:
        cdef size_t n
        cdef base.idxint col0, col1, col2
        # Fast paths for tridiagonal matrices.  These fast paths minimise the
        # number of comparisons and swaps made.
        if size < 2:
            return
        if size == 2:
            if matrix.col_index[ptr] > matrix.col_index[ptr + 1]:
                _sorter_swap(matrix.col_index + ptr, matrix.col_index + ptr+1)
                _sorter_swap(matrix.data + ptr, matrix.data + ptr+1)
            return
        if size == 3:
            # Faster to store rather than re-dereference, and if someone
            # changes the data underneath us, we've got larger problems anyway.
            col0 = matrix.col_index[ptr]
            col1 = matrix.col_index[ptr + 1]
            col2 = matrix.col_index[ptr + 2]
            if col0 < col1:
                if col1 > col2:
                    _sorter_swap(matrix.col_index + ptr+1, matrix.col_index + ptr+2)
                    _sorter_swap(matrix.data + ptr+1, matrix.data + ptr+2)
                    if col0 > col2:
                        _sorter_swap(matrix.col_index + ptr, matrix.col_index + ptr+1)
                        _sorter_swap(matrix.data + ptr, matrix.data + ptr+1)
            elif col1 < col2:
                _sorter_swap(matrix.col_index + ptr, matrix.col_index + ptr+1)
                _sorter_swap(matrix.data + ptr, matrix.data + ptr+1)
                if col0 > col2:
                    _sorter_swap(matrix.col_index + ptr+1, matrix.col_index + ptr+2)
                    _sorter_swap(matrix.data + ptr+1, matrix.data + ptr+2)
            else:
                _sorter_swap(matrix.col_index + ptr, matrix.col_index + ptr+2)
                _sorter_swap(matrix.data + ptr, matrix.data + ptr+2)
            return
        # Now we actually have to do the sort properly.  It's easiest just to
        # copy the data into a temporary structure.
        if size > self.size or self.sort == NULL:
            # realloc(NULL, size) is equivalent to malloc(size), so there's no
            # problem if cols and argsort weren't allocated before.
            self.size = size if size > self.size else self.size
            with gil:
                self.sort = <_data_col *> mem.PyMem_Realloc(self.sort,
                                                            self.size * sizeof(_data_col))
        for n in range(size):
            self.sort[n].data = matrix.data[ptr + n]
            self.sort[n].col = matrix.col_index[ptr + n]
        sort(self.sort, self.sort + size, &_sorter_cmp_struct)
        for n in range(size):
            matrix.data[ptr + n] = self.sort[n].data
            matrix.col_index[ptr + n] = self.sort[n].col

    cdef void copy(self,
                   double complex *dest_data, base.idxint *dest_cols,
                   double complex *src_data, base.idxint *src_cols,
                   size_t size) nogil:
        cdef size_t n, ptr
        # Fast paths for small sizes.  Not pretty, but it speeds things up a
        # lot for up to triadiaongal systems (which are pretty common).
        if size == 0:
            return
        if size == 1:
            dest_cols[0] = src_cols[0]
            dest_data[0] = src_data[0]
            return
        if size == 2:
            if src_cols[0] < src_cols[1]:
                dest_cols[0] = src_cols[0]
                dest_data[0] = src_data[0]
                dest_cols[1] = src_cols[1]
                dest_data[1] = src_data[1]
            else:
                dest_cols[1] = src_cols[0]
                dest_data[1] = src_data[0]
                dest_cols[0] = src_cols[1]
                dest_data[0] = src_data[1]
            return
        if size == 3:
            if src_cols[0] < src_cols[1]:
                if src_cols[0] < src_cols[2]:
                    dest_cols[0] = src_cols[0]
                    dest_data[0] = src_data[0]
                    if src_cols[1] < src_cols[2]:
                        dest_cols[1] = src_cols[1]
                        dest_data[1] = src_data[1]
                        dest_cols[2] = src_cols[2]
                        dest_data[2] = src_data[2]
                    else:
                        dest_cols[2] = src_cols[1]
                        dest_data[2] = src_data[1]
                        dest_cols[1] = src_cols[2]
                        dest_data[1] = src_data[2]
                else:
                    dest_cols[1] = src_cols[0]
                    dest_data[1] = src_data[0]
                    dest_cols[2] = src_cols[1]
                    dest_data[2] = src_data[1]
                    dest_cols[0] = src_cols[2]
                    dest_data[0] = src_data[2]
            elif src_cols[0] < src_cols[2]:
                dest_cols[1] = src_cols[0]
                dest_data[1] = src_data[0]
                dest_cols[0] = src_cols[1]
                dest_data[0] = src_data[1]
                dest_cols[2] = src_cols[2]
                dest_data[2] = src_data[2]
            else:
                dest_cols[2] = src_cols[0]
                dest_data[2] = src_data[0]
                if src_cols[1] < src_cols[2]:
                    dest_cols[0] = src_cols[1]
                    dest_data[0] = src_data[1]
                    dest_cols[1] = src_cols[2]
                    dest_data[1] = src_data[2]
                else:
                    dest_cols[1] = src_cols[1]
                    dest_data[1] = src_data[1]
                    dest_cols[0] = src_cols[2]
                    dest_data[0] = src_data[2]
            return
        # Now we're left with the full case, and we have to sort properly.
        if size > self.size or self.argsort == NULL:
            # realloc(NULL, size) is equivalent to malloc(size), so there's no
            # problem if cols and argsort weren't allocated before.
            self.size = size if size > self.size else self.size
            with gil:
                self.argsort = (
                    <base.idxint **>
                    mem.PyMem_Realloc(self.argsort, self.size * sizeof(base.idxint *))
                )
        # We do the argsort with two levels of indirection to minimise memory
        # allocation and copying requirements when this function is being used
        # to assemble a CSR matrix under an operation which may change the
        # order of the columns (e.g. permute).  First the user makes the
        # columns accessible in some contiguous memory (or if they're not
        # changing them, they can just use the CSR buffers).  We put pointers
        # to each of those columns in the array which actually gets sorted
        # using a comparison function which dereferences the pointers and
        # compares the result.  After the sort, `argsort` will be the pointers
        # sorted according to the new column, and we know that the "lowest"
        # pointer in there has the value `src_cols`, so we can do pointer
        # arithmetic to know which element we should take.
        #
        # This is about 30-40% faster than allocating space for structs of
        # (double complex, idxint), copying in the data and column, sorting and
        # copying into the new arrays.  Allocating the structs actually
        # allocates more space than the pointer method (double complex is
        # very likely to be 2x the size of a pointer, _and_ the struct may need
        # extra padding to be aligned), so it's probably actually worse for
        # cache locality.  Despite the sort relying on pointer dereference in
        # this case, it's actually got very good cache locality.
        for n in range(size):
            self.argsort[n] = src_cols + n
        sort(self.argsort, self.argsort + size, &_sorter_cmp_ptr)
        for n in range(size):
            ptr = self.argsort[n] - src_cols
            dest_cols[n] = src_cols[ptr]
            dest_data[n] = src_data[ptr]

    def __dealloc__(self):
        if self.argsort != NULL:
            mem.PyMem_Free(self.argsort)
        if self.sort != NULL:
            mem.PyMem_Free(self.sort)

cpdef CSR sorted(CSR matrix):
    cdef CSR out = empty_like(matrix)
    cdef Sorter sort
    cdef base.idxint ptr
    cdef size_t row, diff, size=0
    memcpy(out.row_index, matrix.row_index, (matrix.shape[0] + 1) * sizeof(base.idxint))
    for row in range(matrix.shape[0]):
        diff = matrix.row_index[row + 1] - matrix.row_index[row]
        size = diff if diff > size else size
    sort = Sorter(size)
    for row in range(matrix.shape[0]):
        ptr = matrix.row_index[row]
        diff = matrix.row_index[row + 1] - ptr
        sort.copy(out.data + ptr, out.col_index + ptr,
                  matrix.data + ptr, matrix.col_index + ptr,
                  diff)
    return out


cpdef CSR empty(base.idxint rows, base.idxint cols, base.idxint size):
    """
    Allocate an empty CSR matrix of the given shape, with space for `size`
    elements in the `data` and `col_index` arrays.

    This does not initialise any of the memory returned, but sets the last
    element of `row_index` to 0 to indicate that there are 0 non-zero elements.
    """
    if size < 0:
        raise ValueError("size must be a positive integer.")
    # Python doesn't like allocating nothing.
    if size == 0:
        size += 1
    cdef CSR out = CSR.__new__(CSR)
    cdef base.idxint row_size = rows + 1
    out.shape = (rows, cols)
    out.size = size
    out.data =\
        <double complex *> PyDataMem_NEW(size * sizeof(double complex))
    out.col_index =\
        <base.idxint *> PyDataMem_NEW(size * sizeof(base.idxint))
    out.row_index =\
        <base.idxint *> PyDataMem_NEW(row_size * sizeof(base.idxint))
    # Set the number of non-zero elements to 0.
    out.row_index[rows] = 0
    return out


cpdef CSR empty_like(CSR other):
    return empty(other.shape[0], other.shape[1], nnz(other))


cpdef CSR zeros(base.idxint rows, base.idxint cols):
    """
    Allocate the zero matrix with a given shape.  There will not be any room in
    the `data` and `col_index` buffers to add new elements.
    """
    # We always allocate matrices with at least one element to ensure that we
    # actually _are_ asking for memory (Python doesn't like allocating nothing)
    cdef CSR out = empty(rows, cols, 1)
    out.data[0] = out.col_index[0] = 0
    memset(out.row_index, 0, (rows + 1) * sizeof(base.idxint))
    return out


cpdef CSR identity(base.idxint dimension, double complex scale=1):
    """
    Return a square matrix of the specified dimension, with a constant along
    the diagonal.  By default this will be the identity matrix, but if `scale`
    is passed, then the result will be `scale` times the identity.
    """
    cdef CSR out = empty(dimension, dimension, dimension)
    cdef base.idxint i
    for i in range(dimension):
        out.data[i] = scale
        out.col_index[i] = i
        out.row_index[i] = i
    out.row_index[dimension] = dimension
    return out

cpdef CSR from_dense(Dense matrix):
    # Assume worst-case scenario for non-zero.
    cdef CSR out = empty(matrix.shape[0], matrix.shape[1],
                         matrix.shape[0]*matrix.shape[1])
    cdef size_t row, col, ptr_in, ptr_out=0, row_stride, col_stride
    row_stride = 1 if matrix.fortran else matrix.shape[1]
    col_stride = matrix.shape[0] if matrix.fortran else 1
    out.row_index[0] = 0
    for row in range(matrix.shape[0]):
        ptr_in = row_stride * row
        for col in range(matrix.shape[1]):
            if matrix.data[ptr_in] != 0:
                out.data[ptr_out] = matrix.data[ptr_in]
                out.col_index[ptr_out] = col
                ptr_out += 1
            ptr_in += col_stride
        out.row_index[row + 1] = ptr_out
    return out

cdef CSR from_coo_pointers(
    base.idxint *rows, base.idxint *cols, double complex *data,
    base.idxint n_rows, base.idxint n_cols, base.idxint nnz, double tol=0
):
    # Note that COO pointers may not be sorted in row-major order, and that
    # they may contain duplicate entries which should be implicitly summed.
    cdef Accumulator acc = acc_alloc(n_cols)
    cdef CSR out = empty(n_rows, n_cols, nnz)
    cdef double complex *data_tmp
    cdef base.idxint *cols_tmp
    cdef base.idxint row
    cdef size_t ptr_in, ptr_out, ptr_prev
    data_tmp = <double complex *> mem.PyMem_Malloc(nnz * sizeof(double complex))
    cols_tmp = <base.idxint *> mem.PyMem_Malloc(nnz * sizeof(base.idxint))
    if data_tmp == NULL or cols_tmp == NULL:
        raise MemoryError
    with nogil:
        memset(out.row_index, 0, (n_rows + 1) * sizeof(base.idxint))
        for ptr_in in range(nnz):
            out.row_index[rows[ptr_in] + 1] += 1
        for ptr_out in range(n_rows):
            out.row_index[ptr_out + 1] += out.row_index[ptr_out]
        # out.row_index is currently in the normal output form, but we're
        # temporarily going to modify it to keep track of how many values we've
        # placed in each row as we iterate through.  At every state,
        # out.row_index[row] will contain a pointer to the next location that
        # an element should be placed in this row.
        for ptr_in in range(nnz):
            row = rows[ptr_in]
            ptr_out = out.row_index[row]
            cols_tmp[ptr_out] = cols[ptr_in]
            data_tmp[ptr_out] = data[ptr_in]
            out.row_index[row] += 1
        # Apply the scatter/gather pattern to find the actual number of
        # non-zero elements we're writing into each row (since there's a sum,
        # there may be some zeros of duplicates).  Remember we also need to
        # shift the row_index array back to what it was before as well.
        ptr_out = 0
        ptr_prev = 0
        for row in range(n_rows):
            for ptr_in in range(ptr_prev, out.row_index[row]):
                acc_scatter(&acc, data_tmp[ptr_in], cols_tmp[ptr_in])
            ptr_prev = out.row_index[row]
            out.row_index[row] = ptr_out
            ptr_out += acc_gather(&acc, out.data + ptr_out, out.col_index + ptr_out, tol)
            acc_reset(&acc)
        out.row_index[n_rows] = ptr_out
    mem.PyMem_Free(data_tmp)
    mem.PyMem_Free(cols_tmp)
    acc_free(&acc)
    return out


cdef inline base.idxint _diagonal_length(
    base.idxint offset, base.idxint n_rows, base.idxint n_cols,
) nogil:
    if offset > 0:
        return n_rows if offset <= n_cols - n_rows else n_cols - offset
    return n_cols if offset > n_cols - n_rows else n_rows + offset

cdef CSR diag(
    double complex[:] diagonal, base.idxint offset,
    base.idxint n_rows, base.idxint n_cols,
):
    """
    Construct a CSR matrix with a single non-zero diagonal.

    Parameters
    ----------
    diagonal : indexable of double complex
        The entries (including zeros) that should be placed on the diagonal in
        the output matrix.  Each entry must have enough entries in it to fill
        the relevant diagonal.

    offsets : idxint
        The index of the diagonals.  An offset of 0 is the main diagonal,
        positive values are above the main diagonal and negative ones are below
        the main diagonal.

    n_rows, n_cols : idxint
        The shape of the output.  The result does not need to be square, but
        the diagonal must be of the correct length to fit in.
    """
    if n_rows < 0 or n_cols < 0:
        raise ValueError("shape must be positive")
    cdef base.idxint nnz = len(diagonal)
    cdef base.idxint n_diag = _diagonal_length(offset, n_rows, n_cols)
    if nnz != n_diag:
        raise ValueError("incorrect number of diagonal elements")
    cdef CSR out = empty(n_rows, n_cols, nnz)
    cdef base.idxint start_row = 0 if offset >= 0 else -offset
    cdef base.idxint col = 0 if offset <= 0 else offset
    memset(out.row_index, 0, (start_row + 1) * sizeof(base.idxint))
    nnz = 0
    for row in range(start_row + 1, start_row + n_diag + 1):
        out.col_index[nnz] = col
        out.data[nnz] = diagonal[nnz]
        col += 1
        nnz += 1
        out.row_index[row] = nnz
    for row in range(start_row + n_diag + 1, n_rows + 1):
        out.row_index[row] = nnz
    return out

cdef CSR diags_(
    list diagonals, base.idxint[:] offsets,
    base.idxint n_rows, base.idxint n_cols,
):
    """
    Construct a CSR matrix from a list of diagonals and their offsets.  The
    offsets are assumed to be in sorted order.  This is the C-only interface to
    csr.diags, and inputs are not sanity checked (use the Python interface for
    that).

    Parameters
    ----------
    diagonals : list of indexable of double complex
        The entries (including zeros) that should be placed on the diagonals in
        the output matrix.  Each entry must have enough entries in it to fill
        the relevant diagonal (not checked).

    offsets : idxint[:]
        The indices of the diagonals.  These should be sorted and without
        duplicates.  `offsets[i]` is the location of the values `diagonals[i]`.
        An offset of 0 is the main diagonal, positive values are above the main
        diagonal and negative ones are below the main diagonal.

    n_rows, n_cols : idxint
        The shape of the output.  The result does not need to be square, but
        the diagonals must be of the correct length to fit in.
    """
    cdef size_t n_diagonals = len(diagonals)
    if n_diagonals == 0:
        return zeros(n_rows, n_cols)
    cdef base.idxint k, row, start_row, offset, nnz=0,
    cdef base.idxint min_k=n_diagonals, max_k=n_diagonals
    cdef double complex value
    for k in range(n_diagonals):
        offset = offsets[k]
        if offset >= 0 and min_k > k:
            min_k = k
        nnz += _diagonal_length(offset, n_rows, n_cols)
    cdef CSR out = empty(n_rows, n_cols, nnz)
    nnz = 0
    out.row_index[0] = 0
    for row in range(n_rows):
        if min_k > 0:
            offset = offsets[min_k - 1]
            start_row = 0 if offset >= 0 else -offset
            if start_row == row:
                min_k -= 1
        if max_k > 0:
            offset = offsets[max_k - 1]
            start_row = 0 if offset >= 0 else -offset
            if start_row + _diagonal_length(offset, n_rows, n_cols) - 1 < row:
                max_k -= 1
        for k in range(min_k, max_k):
            offset = offsets[k]
            value = diagonals[k][row if offset >= 0 else row + offset]
            if value == 0:
                continue
            out.data[nnz] = value
            out.col_index[nnz] = row + offset
            nnz += 1
        out.row_index[row + 1] = nnz
    return out

@cython.wraparound(True)
def diags(diagonals, offsets=None, shape=None):
    """
    Construct a CSR matrix from diagonals and their offsets.  Using this
    function in single-argument form produces a square matrix with the given
    values on the main diagonal.

    With lists of diagonals and offsets, the matrix will be the smallest
    possible square matrix if shape is not given, but in all cases the
    diagonals must fit exactly with no extra or missing elements.  Duplicated
    diagonals will be summed together in the output.

    Parameters
    ----------
    diagonals : sequence of array_like of complex or array_like of complex
        The entries (including zeros) that should be placed on the diagonals in
        the output matrix.  Each entry must have enough entries in it to fill
        the relevant diagonal and no more.

    offsets : sequence of integer or integer, optional
        The indices of the diagonals.  `offsets[i]` is the location of the
        values `diagonals[i]`.  An offset of 0 is the main diagonal, positive
        values are above the main diagonal and negative ones are below the main
        diagonal.

    shape : tuple, optional
        The shape of the output as (``rows``, ``columns``).  The result does
        not need to be square, but the diagonals must be of the correct length
        to fit in exactly.
    """
    cdef base.idxint n_rows, n_cols, offset
    try:
        diagonals = list(diagonals)
        if diagonals and np.isscalar(diagonals[0]):
            # Catch the case where we're being called as (for example)
            #   diags([1, 2, 3], 0)
            # with a single diagonal and offset.
            diagonals = [diagonals]
    except TypeError:
        raise TypeError("diagonals must be a list of arrays of complex") from None
    if offsets is None:
        if len(diagonals) == 0:
            offsets = []
        elif len(diagonals) == 1:
            offsets = [0]
        else:
            raise TypeError("offsets must be supplied if passing more than one diagonal")
    offsets = np.atleast_1d(offsets)
    if offsets.ndim > 1:
        raise ValueError("offsets must be a 1D array of integers")
    if len(diagonals) != len(offsets):
        raise ValueError("number of diagonals does not match number of offsets")
    if len(diagonals) == 0:
        if shape is None:
            raise ValueError("cannot construct matrix with no diagonals without a shape")
        else:
            n_rows, n_cols = shape
        return zeros(n_rows, n_cols)
    order = np.argsort(offsets)
    diagonals_ = []
    offsets_ = []
    prev, cur = None, None
    for i in order:
        cur = offsets[i]
        if cur == prev:
            diagonals_[-1] += np.asarray(diagonals[i], dtype=np.complex128)
        else:
            offsets_.append(cur)
            diagonals_.append(np.asarray(diagonals[i], dtype=np.complex128))
        prev = cur
    if shape is None:
        n_rows = n_cols = abs(offsets_[0]) + len(diagonals_[0])
    else:
        try:
            n_rows, n_cols = shape
        except (TypeError, ValueError):
            raise TypeError("shape must be a 2-tuple of positive integers")
        if n_rows < 0 or n_cols < 0:
            raise ValueError("shape must be a 2-tuple of positive integers")
    for i in range(len(diagonals_)):
        offset = offsets_[i]
        if len(diagonals_[i]) != _diagonal_length(offset, n_rows, n_cols):
            raise ValueError("given diagonals do not have the correct lengths")
    if n_rows == 0 and n_cols == 0:
        raise ValueError("can't produce a 0x0 matrix")
    if len(offsets) == 1:
        # Fast path for a single diagonal.
        return diag(diagonals_[0], offsets_[0], n_rows, n_cols)
    return diags_(
        diagonals_, np.array(offsets_, dtype=idxint_dtype), n_rows, n_cols,
    )



# This module exists to supply a couple of very standard constant matrices
# which are used in the data layer, and within `Qobj` itself.  Other matrices
# (e.g. `create`) should not be here, but should be defined within the
# higher-level components of QuTiP instead.

from . import csr, dense
from .csr import CSR
from .dense import Dense
from .dispatch import Dispatcher as _Dispatcher
import inspect as _inspect

__all__ = ['zeros', 'identity']

zeros = _Dispatcher(
    _inspect.Signature([
        _inspect.Parameter('rows', _inspect.Parameter.POSITIONAL_OR_KEYWORD),
        _inspect.Parameter('cols', _inspect.Parameter.POSITIONAL_OR_KEYWORD),
    ]),
    name='zeros',
    module=__name__,
    inputs=(),
    out=True,
)
zeros.__doc__ =\
    """
    Create matrix representation of 0 with the given dimensions.

    Depending on the selected output type, this may or may not actually
    contained explicit values; sparse matrices will typically contain nothing
    (which is their representation of 0), and dense matrices will still be
    filled.

    Parameters
    ----------
    rows, cols : int
        The number of rows and columns in the output matrix.
    """
zeros.add_specialisations([
    (CSR, csr.zeros),
    (Dense, dense.zeros),
], _defer=True)

identity = _Dispatcher(
    _inspect.Signature([
        _inspect.Parameter('dimension',
                           _inspect.Parameter.POSITIONAL_OR_KEYWORD),
        _inspect.Parameter('scale', _inspect.Parameter.POSITIONAL_OR_KEYWORD,
                           default=1),
    ]),
    name='identity',
    module=__name__,
    inputs=(),
    out=True,
)
identity.__doc__ =\
    """
    Create a square identity matrix of the given dimension.  Optionally, the
    `scale` can be given, where all the diagonal elements will be that instead
    of 1.

    Parameters
    ----------
    dimension : int
        The dimension of the square output identity matrix.
    scale : complex, optional
        The element which should be placed on the diagonal.
    """
identity.add_specialisations([
    (CSR, csr.identity),
    (Dense, dense.identity),
], _defer=True)

del _Dispatcher, _inspect



"""Module replicating the qutip_qip package from within qutip."""
import sys

try:
    import qutip_qip
    del qutip_qip
    sys.modules["qutip.qip"] = sys.modules["qutip_qip"]
except ImportError:
    raise ImportError(
        "Importing 'qutip.qip' requires the 'qutip_qip' package. Install it "
        "with `pip install qutip-qip` (for more details, go to "
        "https://qutip-qip.readthedocs.io/)."
    )



from qutip.core.data import CSR, Data, csr, Dense
import qutip.core.data as _data
import scipy.sparse.linalg as splinalg
import numpy as np
from qutip.settings import settings
if settings.has_mkl:
    from qutip._mkl.spsolve import mkl_spsolve
else:
    mkl_spsolve = None


def _splu(A, B, **kwargs):
    lu = splinalg.splu(A, **kwargs)
    return lu.solve(B)


def solve_csr_dense(matrix: CSR, target: Dense, method=None,
                    options: dict={}) -> Dense:
    """
    Solve ``Ax=b`` for ``x``.

    Parameters:
    -----------

    matrix : CSR
        The matrix ``A``.

    target : Data
        The matrix or vector ``b``.

    method : str {"spsolve", "splu", "mkl_spsolve", etc.}, default="spsolve"
        The function to use to solve the system. Any function from
        scipy.sparse.linalg which solve the equation Ax=b can be used.
        `splu` from the same and `mkl_spsolve` are also valid choice.

    options : dict
        Keywork options to pass to the solver. Refer to the documenentation in
        scipy.sparse.linalg of the used method for a list of supported keyword.
        The keyword "csc" can be set to ``True`` to convert the sparse matrix
        before passing it to the solver.

    .. note::
        Options for ``mkl_spsolve`` are presently only found in the source
        code.

    Returns:
    --------
    x : Dense
        Solution to the system Ax = b.
    """
    if matrix.shape[0] != matrix.shape[1]:
        raise ValueError("can only solve using square matrix")
    if matrix.shape[1] != target.shape[0]:
        raise ValueError("target does not match the system")

    b = target.as_ndarray()

    method = method or "spsolve"

    if method == "splu":
        solver = _splu
    elif hasattr(splinalg, method):
        solver = getattr(splinalg, method)
    elif method == "mkl_spsolve" and mkl_spsolve is None:
        raise ValueError("mkl is not available")
    elif method == "mkl_spsolve":
        solver = mkl_spsolve
    else:
        raise ValueError(f"Unknown sparse solver {method}.")

    options = options.copy()
    M = matrix.as_scipy()
    if options.pop("csc", False):
        M = M.tocsc()

    out = solver(M, b, **options)

    if isinstance(out, tuple) and len(out) == 2:
        # iterative method return a success flag
        out, check = out
        if check == 0:
            # Successful
            pass
        elif check > 0:
            raise RuntimeError(
                f"scipy.sparse.linalg.{method} error: Tolerance was not"
                f" reached. Error code: {check}"
            )

        elif check < 0:
            raise RuntimeError(
                f"scipy.sparse.linalg.{method} error: Bad input. "
                f"Error code: {check}"
            )
    elif isinstance(out, tuple) and len(out) > 2:
        # least sqare method return residual, flag, etc.
        out, *_ = out
    return Dense(out, copy=False)


def solve_dense(matrix: Dense, target: Data, method=None,
                options: dict={}) -> Dense:
    """
    Solve ``Ax=b`` for ``x``.

    Parameters:
    -----------

    matrix : Dense
        The matrix ``A``.

    target : Data
        The matrix or vector ``b``.

    method : str {"solve", "lstsq"}, default="solve"
        The function from numpy.linalg to use to solve the system.

    options : dict
        Options to pass to the solver. "lstsq" use "rcond" while, "solve" do
        not use any.

    Returns:
    --------
    x : Dense
        Solution to the system Ax = b.
    """
    if matrix.shape[0] != matrix.shape[1]:
        raise ValueError("can only solve using square matrix")
    if matrix.shape[1] != target.shape[0]:
        raise ValueError("target does not match the system")

    if isinstance(target, Dense):
        b = target.as_ndarray()
    else:
        b = target.to_array()

    if method in ["solve", None]:
        out = np.linalg.solve(matrix.as_ndarray(), b)
    elif method == "lstsq":
        out, *_ = np.linalg.lstsq(
            matrix.as_ndarray(),
            b,
            rcond=options.get("rcond", None)
        )
    else:
        raise ValueError(f"Unknown solver {method},"
                         " 'solve' and 'lstsq' are supported.")
    return Dense(out, copy=False)


from .dispatch import Dispatcher as _Dispatcher
import inspect as _inspect


solve = _Dispatcher(
    _inspect.Signature([
        _inspect.Parameter('matrix', _inspect.Parameter.POSITIONAL_OR_KEYWORD),
        _inspect.Parameter('target', _inspect.Parameter.POSITIONAL_OR_KEYWORD),
        _inspect.Parameter('method', _inspect.Parameter.POSITIONAL_OR_KEYWORD,
                           default=None),
        _inspect.Parameter('options', _inspect.Parameter.POSITIONAL_OR_KEYWORD,
                           default={}),
    ]),
    name='solve',
    module=__name__,
    inputs=('matrix', 'target'),
    out=True,
)
solve.__doc__ = """
    Solve ``Ax=b`` for ``x``.

    Parameters:
    -----------
    matrix : Data
    The matrix ``A``.

    target : Data
    The matrix or vector ``b``.

    method : str
        The function to use to solve the system. Function which solve the
        equation Ax=b from scipy.sparse.linalg (CSR ``matrix``) or
        numpy.linalg (Dense ``matrix``) can be used.
        Sparse cases also accept `splu` and `mkl_spsolve`.

    options : dict
        Keywork options to pass to the solver. Refer to the documenentation
        for the chosen method in scipy.sparse.linalg or numpy.linalg.
        The keyword "csc" can be set to ``True`` to convert the sparse matrix
        in sparse cases.

    .. note::
        Options for ``mkl_spsolve`` are presently only found in the source
        code.

    Returns:
    --------
    x : Data
        Solution to the system Ax = b.
"""
solve.add_specialisations([
    (CSR, Dense, Dense, solve_csr_dense),
    (Dense, Dense, Dense, solve_dense),
], _defer=True)


del _Dispatcher
del _inspect



"""
This module includes a collection of testing functions for the QuTiP scattering
module. Tests are approximate with low resolution to minimize runtime.
"""

# Author:  Ben Bartlett
# Contact: benbartlett@stanford.edu

import numpy as np
from numpy.testing import assert_, run_module_suite
from qutip import create, destroy, basis
from qutip.solver.scattering import *


class TestScattering:
    """
    A test class for the QuTiP quantum optical scattering module. These tests
    only use the two-level system for comparison, since larger systems can
    take a long time to run.
    """

    def testScatteringProbability(self):
        """
        Asserts that pi pulse in TLS has P0 ~ 0 and P0+P1+P2 ~ 1
        """
        w0 = 1.0 * 2 * np.pi
        gamma = 1.0
        sm = np.sqrt(gamma) * destroy(2)
        pulseArea = np.pi
        pulseLength = 0.2 / gamma
        RabiFreq = pulseArea / (2 * pulseLength)
        psi0 = basis(2, 0)
        tlist = np.geomspace(gamma, 10 * gamma, 40) - gamma
        # Define TLS Hamiltonian
        H0S = w0 * create(2) * destroy(2)
        H1S1 = lambda t, args: \
            RabiFreq * 1j * np.exp(-1j * w0 * t) * (t < pulseLength)
        H1S2 = lambda t, args: \
            RabiFreq * -1j * np.exp(1j * w0 * t) * (t < pulseLength)
        Htls = [H0S, [sm.dag(), H1S1], [sm, H1S2]]
        # Run the test
        P0 = scattering_probability(Htls, psi0, 0, [sm], tlist)
        P1 = scattering_probability(Htls, psi0, 1, [sm], tlist)
        P2 = scattering_probability(Htls, psi0, 2, [sm], tlist)
        assert P0 < 1e-3
        assert np.abs(P0 + P1 + P2 - 1) < 1e-3

    def testScatteringAmplitude(self):
        """
        Asserts that a 2pi pulse in TLS has ~0 amplitude after pulse
        """
        w0 = 1.0 * 2 * np.pi
        gamma = 1.0
        sm = np.sqrt(gamma) * destroy(2)
        pulseArea = 2 * np.pi
        pulseLength = 0.2 / gamma
        RabiFreq = pulseArea / (2 * pulseLength)
        psi0 = basis(2, 0)
        T = 50
        tlist = np.linspace(0, 1 / gamma, T)
        # Define TLS Hamiltonian
        H0S = w0 * create(2) * destroy(2)
        H1S1 = lambda t, args: \
            RabiFreq * 1j * np.exp(-1j * w0 * t) * (t < pulseLength)
        H1S2 = lambda t, args: \
            RabiFreq * -1j * np.exp(1j * w0 * t) * (t < pulseLength)
        Htls = [H0S, [sm.dag(), H1S1], [sm, H1S2]]
        # Run the test
        state = temporal_scattered_state(Htls, psi0, 1, [sm], tlist)
        basisVec = temporal_basis_vector([[40]], T)
        assert np.abs(basisVec.dag() @ state) < 1e-3

    def testWaveguideSplit(self):
        """
        Checks that a trivial splitting of a waveguide collapse operator like
        [sm] -> [sm/sqrt2, sm/sqrt2] doesn't affect the normalization or result
        """
        gamma = 1.0
        sm = np.sqrt(gamma) * destroy(2)
        pulseArea = np.pi
        pulseLength = 0.2 / gamma
        RabiFreq = pulseArea / (2 * pulseLength)
        psi0 = basis(2, 0)
        tlist = np.geomspace(gamma, 10 * gamma, 40) - gamma
        # Define TLS Hamiltonian with rotating frame transformation
        Htls = [[sm.dag() + sm, lambda t, args: RabiFreq * (t < pulseLength)]]
        # Run the test
        c_ops = [sm]
        c_ops_split = [sm / np.sqrt(2), sm / np.sqrt(2)]
        P1 = scattering_probability(Htls, psi0, 1, c_ops, tlist)
        P1_split = scattering_probability(Htls, psi0, 1, c_ops_split, tlist)
        assert np.abs(P1/P1_split - 1) < 1e-7



__all__ = ["Options", "SolverOptions"]

import warnings


def SolverOptions(*args, **kwargs):
    warnings.warn(
        "Dedicated options class are no longer needed, "
        "options should be passed as dict to solvers.",
        FutureWarning
    )
    return kwargs


def Options(**kwargs):
    warnings.warn(
        "Dedicated options class are no longer needed, "
        "options should be passed as dict to solvers.",
        FutureWarning
    )
    return kwargs


class _SolverOptions(dict):
    """
    Class to hold options for solver and integrator.

    Parameters
    ----------
    default : dict
        Default dict, only keys in this will be accepted.
    feedback : callable, ``f(keys : set) -> None``, optional
        Function to called when an item is updated.
    name : str, optional
        Name of the solver or integrator that use this. Used in __repr__ only.
    doc : str, optional
        Overwrite the __doc__ of the instance.
    """
    def __init__(
        self, default, feedback=None, name="", doc="", /, **kwargs
    ):
        self._default = default
        self.__doc__ = doc
        self._feedback = feedback
        self._name = name
        extra_keys = kwargs.keys() - default.keys()
        if extra_keys:
            raise KeyError(f"Options {extra_keys} are not supported.")
        super().__init__(**{**self._default, **kwargs})

    def __setitem__(self, key, val):
        if key not in self._default:
            raise KeyError(f"Options {key} is not supported.")
        if val is None:
            val = self._default[key]
        if val == self[key]:
            return
        super().__setitem__(key, val)
        if self._feedback:
            self._feedback(key)

    def __delitem__(self, key):
        if key not in self._default:
            raise KeyError(f"Options {key} is not supported.")
        super().__setitem__(key, self._default[key])
        if self._feedback:
            self._feedback(key)

    def copy(self):
        return self.__class__(
            self._default,
            self._feedback,
            self._name,
            self.__doc__,
            **self
        )

    def __str__(self):
        lines = []
        longest = max(len(key) for key in self.keys())
        lines.append(f"Options for {self._name}:")
        for key in self.keys():
            default = "(default)" if self[key] == self._default[key] else ""
            lines.append(f"    {key:{longest}} : "
                         f"{self[key].__repr__():{70-longest}}"
                         f"{default}")
        return "\n".join(lines)

    @classmethod
    def _from_reduced(cls, default, feedback, name, doc, keys, args):
        return cls(default, feedback, name, doc, **{
            key: arg for key, arg in zip(keys, args)
        })

    def __reduce__(self):
        return (
            self._from_reduced,
            (
                self._default,
                self._feedback,
                self._name,
                self.__doc__,
                tuple(self.keys()),
                tuple(self.values())
                )
            )



__all__ = ['mcsolve', "MCSolver"]

import warnings

import numpy as np
from copy import copy
from ..core import QobjEvo, spre, spost, Qobj, unstack_columns, liouvillian
from .multitraj import MultiTrajSolver
from .solver_base import Solver
from .result import McResult, Result
from .mesolve import mesolve, MESolver
import qutip.core.data as _data
from time import time


def mcsolve(H, state, tlist, c_ops=(), e_ops=None, ntraj=500, *,
            args=None, options=None, seeds=None, target_tol=None, timeout=None):
    r"""
    Monte Carlo evolution of a state vector :math:`|\psi \rangle` for a
    given Hamiltonian and sets of collapse operators. Options for the
    underlying ODE solver are given by the Options class.

    Parameters
    ----------
    H : :class:`qutip.Qobj`, :class:`qutip.QobjEvo`, ``list``, callable.
        System Hamiltonian as a Qobj, QobjEvo. It can also be any input type
        that QobjEvo accepts (see :class:`qutip.QobjEvo`'s documentation).
        ``H`` can also be a superoperator (liouvillian) if some collapse
        operators are to be treated deterministically.

    state : :class:`qutip.Qobj`
        Initial state vector.

    tlist : array_like
        Times at which results are recorded.

    c_ops : list
        A ``list`` of collapse operators in any input type that QobjEvo accepts
        (see :class:`qutip.QobjEvo`'s documentation). They must be operators
        even if ``H`` is a superoperator. If none are given, the solver will
        defer to ``sesolve`` or ``mesolve``.

    e_ops : list, [optional]
        A ``list`` of operator as Qobj, QobjEvo or callable with signature of
        (t, state: Qobj) for calculating expectation values. When no ``e_ops``
        are given, the solver will default to save the states.

    ntraj : int
        Maximum number of trajectories to run. Can be cut short if a time limit
        is passed with the ``timeout`` keyword or if the target tolerance is
        reached, see ``target_tol``.

    args : None / dict
        Arguments for time-dependent Hamiltonian and collapse operator terms.

    options : None / dict
        Dictionary of options for the solver.

        - store_final_state : bool [False]
          Whether or not to store the final state of the evolution in the
          result class.
        - store_states : bool, NoneType, [None]
          Whether or not to store the state vectors or density matrices.
          On `None` the states will be saved if no expectation operators are
          given.
        - progress_bar : str {'text', 'enhanced', 'tqdm', ''}, ['text']
          How to present the solver progress.
          'tqdm' uses the python module of the same name and raise an error
          if not installed. Empty string or False will disable the bar.
        - progress_kwargs : dict, [{"chunk_size": 10}]
          kwargs to pass to the progress_bar. Qutip's bars use `chunk_size`.
        - method : str {"adams", "bdf", "dop853", "vern9", etc.} ["adams"]
          Which differential equation integration method to use.
        - keep_runs_results : bool, [False]
          Whether to store results from all trajectories or just store the
          averages.
        - map : str {"serial", "parallel", "loky"}, ["serial"]
          How to run the trajectories. "parallel" uses concurent module to run
          in parallel while "loky" use the module of the same name to do so.
        - job_timeout : NoneType, int, [None]
          Maximum time to compute one trajectory.
        - num_cpus : NoneType, int, [None]
          Number of cpus to use when running in parallel. ``None`` detect the
          number of available cpus.
        - norm_t_tol, norm_tol, norm_steps : float, float, int, [1e-6, 1e-4, 5]
          Parameters used to find the collapse location. ``norm_t_tol`` and
          ``norm_tol`` are the tolerance in time and norm respectively.
          An error will be raised if the collapse could not be found within
          ``norm_steps`` tries.
        - mc_corr_eps : float, [1e-10]
          Small number used to detect non-physical collapse caused by numerical
          imprecision.
        - atol, rtol : float, [1e-8, 1e-6]
          Absolute and relative tolerance of the ODE integrator.
        - nsteps : int [2500]
          Maximum number of (internally defined) steps allowed in one ``tlist``
          step.
        - max_step : float, [0]
          Maximum lenght of one internal step. When using pulses, it should be
          less than half the width of the thinnest pulse.

    seeds : int, SeedSequence, list, [optional]
        Seed for the random number generator. It can be a single seed used to
        spawn seeds for each trajectory or a list of seeds, one for each
        trajectory. Seeds are saved in the result and they can be reused with::

            seeds=prev_result.seeds

    target_tol : {float, tuple, list}, optional
        Target tolerance of the evolution. The evolution will compute
        trajectories until the error on the expectation values is lower than
        this tolerance. The maximum number of trajectories employed is
        given by ``ntraj``. The error is computed using jackknife resampling.
        ``target_tol`` can be an absolute tolerance or a pair of absolute and
        relative tolerance, in that order. Lastly, it can be a list of pairs of
        (atol, rtol) for each e_ops.

    timeout : float [optional]
        Maximum time for the evolution in second. When reached, no more
        trajectories will be computed. Overwrite the option of the same name.

    Returns
    -------
    results : :class:`qutip.solver.Result`
        Object storing all results from the simulation. Which results is saved
        depends on the presence of ``e_ops`` and the options used. ``collapse``
        and ``photocurrent`` is available to Monte Carlo simulation results.

    .. note:
        The simulation will end when the first end condition is reached between
        ``ntraj``, ``timeout`` and ``target_tol``.
    """
    H = QobjEvo(H, args=args, tlist=tlist)
    if not isinstance(c_ops, (list, tuple)):
        c_ops = [c_ops]
    c_ops = [QobjEvo(c_op, args=args, tlist=tlist) for c_op in c_ops]

    if len(c_ops) == 0:
        if options is None:
            options = {}
        options = {
            key: options[key]
            for key in options
            if key in MESolver.solver_options
        }
        return mesolve(H, state, tlist, e_ops=e_ops, args=args,
                       options=options)

    if isinstance(ntraj, (list, tuple)):
        raise TypeError(
            "ntraj must be an integer. "
            "A list of numbers is not longer supported."
        )

    mc = MCSolver(H, c_ops, options=options)
    result = mc.run(state, tlist=tlist, ntraj=ntraj, e_ops=e_ops,
                    seed=seeds, target_tol=target_tol, timeout=timeout)
    return result


class MCIntegrator:
    """
    Integrator like object for mcsolve trajectory.
    """
    name = "mcsolve"

    def __init__(self, integrator, c_ops, n_ops, options=None):
        self._integrator = integrator
        self._c_ops = c_ops
        self._n_ops = n_ops
        self.options = options
        self._generator = None
        self.method = f"{self.name} {self._integrator.method}"
        self._is_set = False
        self.issuper = c_ops[0].issuper

    def set_state(self, t, state0, generator):
        """
        Set the state of the ODE solver.

        Parameters
        ----------
        t : float
            Initial time

        state0 : qutip.Data
            Initial state.

        generator : numpy.random.generator
            Random number generator.
        """
        self.collapses = []
        self._generator = generator
        self.target_norm = self._generator.random()
        self._integrator.set_state(t, state0)
        self._is_set = True

    def get_state(self, copy=True):
        return *self._integrator.get_state(copy), self._generator

    def integrate(self, t, copy=False):
        t_old, y_old = self._integrator.get_state(copy=False)
        norm_old = self._prob_func(y_old)
        while t_old < t:
            t_step, state = self._integrator.mcstep(t, copy=False)
            norm = self._prob_func(state)
            if norm <= self.target_norm:
                t_col, state = self._find_collapse_time(norm_old, norm,
                                                        t_old, t_step)
                self._do_collapse(t_col, state)
                t_old, y_old = self._integrator.get_state(copy=False)
                norm_old = 1.
            else:
                t_old, y_old = t_step, state
                norm_old = norm

        return t_old, _data.mul(y_old, 1 / self._norm_func(y_old))

    def run(self, tlist):
        for t in tlist[1:]:
            yield self.integrate(t, False)

    def reset(self, hard=False):
        self._integrator.reset(hard)

    def _prob_func(self, state):
        if self.issuper:
            return _data.norm.trace(unstack_columns(state))
        return _data.norm.l2(state)**2

    def _norm_func(self, state):
        if self.issuper:
            return _data.norm.trace(unstack_columns(state))
        return _data.norm.l2(state)

    def _find_collapse_time(self, norm_old, norm, t_prev, t_final):
        """Find the time of the collapse and state just before it."""
        tries = 0
        while tries < self.options['norm_steps']:
            tries += 1
            if (t_final - t_prev) < self.options['norm_t_tol']:
                t_guess = t_final
                _, state = self._integrator.get_state()
                break
            t_guess = (
                t_prev
                + ((t_final - t_prev)
                   * np.log(norm_old / self.target_norm)
                   / np.log(norm_old / norm))
            )
            if (t_guess - t_prev) < self.options['norm_t_tol']:
                t_guess = t_prev + self.options['norm_t_tol']
            _, state = self._integrator.mcstep(t_guess, copy=False)
            norm2_guess = self._prob_func(state)
            if (
                np.abs(self.target_norm - norm2_guess) <
                self.options['norm_tol'] * self.target_norm
            ):
                break
            elif (norm2_guess < self.target_norm):
                # t_guess is still > t_jump
                t_final = t_guess
                norm = norm2_guess
            else:
                # t_guess < t_jump
                t_prev = t_guess
                norm_old = norm2_guess

        if tries >= self.options['norm_steps']:
            raise RuntimeError(
                "Could not find the collapse time within desired tolerance. "
                "Increase accuracy of the ODE solver or lower the tolerance "
                "with the options 'norm_steps', 'norm_tol', 'norm_t_tol'.")

        return t_guess, state

    def _do_collapse(self, collapse_time, state):
        """
        Do the collapse:
        - Find which operator did the collapse.
        - Update the state and Integrator.
        - Next collapse norm location
        - Store collapse info.
        """
        # collapse_time, state is at the collapse
        if len(self._n_ops) == 1:
            which = 0
        else:
            probs = np.zeros(len(self._n_ops))
            for i, n_op in enumerate(self._n_ops):
                probs[i] = n_op.expect_data(collapse_time, state).real
            probs = np.cumsum(probs)
            which = np.searchsorted(probs,
                                    probs[-1] * self._generator.random())

        state_new = self._c_ops[which].matmul_data(collapse_time, state)
        new_norm = self._norm_func(state_new)
        if new_norm < self.options['mc_corr_eps']:
            # This happen when the collapse is caused by numerical error
            state_new = _data.mul(state, 1 / self._norm_func(state))
        else:
            state_new = _data.mul(state_new, 1 / new_norm)
            self.collapses.append((collapse_time, which))
            self.target_norm = self._generator.random()
        self._integrator.set_state(collapse_time, state_new)

    def arguments(self, args):
        if args:
            self._integrator.arguments(args)
            for c_op in self._c_ops:
                c_op.arguments(args)
            for n_op in self._n_ops:
                n_op.arguments(args)

    @property
    def integrator_options(self):
        return self._integrator.integrator_options


# -----------------------------------------------------------------------------
# MONTE CARLO CLASS
# -----------------------------------------------------------------------------
class MCSolver(MultiTrajSolver):
    r"""
    Monte Carlo Solver of a state vector :math:`|\psi \rangle` for a
    given Hamiltonian and sets of collapse operators. Options for the
    underlying ODE solver are given by the Options class.

    Parameters
    ----------
    H : :class:`qutip.Qobj`, :class:`qutip.QobjEvo`, ``list``, callable.
        System Hamiltonian as a Qobj, QobjEvo. It can also be any input type
        that QobjEvo accepts (see :class:`qutip.QobjEvo`'s documentation).
        ``H`` can also be a superoperator (liouvillian) if some collapse
        operators are to be treated deterministically.

    c_ops : list
        A ``list`` of collapse operators in any input type that QobjEvo accepts
        (see :class:`qutip.QobjEvo`'s documentation). They must be operators
        even if ``H`` is a superoperator.

    options : SolverOptions, [optional]
        Options for the evolution.

    seed : int, SeedSequence, list, [optional]
        Seed for the random number generator. It can be a single seed used to
        spawn seeds for each trajectory or a list of seed, one for each
        trajectory. Seeds are saved in the result and can be reused with::
            seeds=prev_result.seeds
    """
    name = "mcsolve"
    resultclass = McResult
    solver_options = {
        "progress_bar": "text",
        "progress_kwargs": {"chunk_size": 10},
        "store_final_state": False,
        "store_states": None,
        "keep_runs_results": False,
        "method": "adams",
        "map": "serial",
        "job_timeout": None,
        "num_cpus": None,
        "bitgenerator": None,
        "mc_corr_eps": 1e-10,
        "norm_steps": 5,
        "norm_t_tol": 1e-6,
        "norm_tol": 1e-4,
    }

    def __init__(self, H, c_ops, *, options=None):
        _time_start = time()

        if isinstance(c_ops, (Qobj, QobjEvo)):
            c_ops = [c_ops]
        c_ops = [QobjEvo(c_op) for c_op in c_ops]

        if H.issuper:
            self._c_ops = [
                spre(c_op) * spost(c_op.dag()) if c_op.isoper else c_op
                for c_op in c_ops
            ]
            self._n_ops = self._c_ops
            rhs = QobjEvo(H)
            for c_op in c_ops:
                cdc = c_op.dag() @ c_op
                rhs -= 0.5 * (spre(cdc) + spost(cdc))
        else:
            self._c_ops = c_ops
            self._n_ops = [c_op.dag() * c_op for c_op in c_ops]
            rhs = -1j * QobjEvo(H)
            for n_op in self._n_ops:
                rhs -= 0.5 * n_op

        self._num_collapse = len(self._c_ops)

        super().__init__(rhs, options=options)

    def _restore_state(self, data, *, copy=True):
        """
        Retore the Qobj state from its data.
        """
        if self._state_metadata['dims'] == self.rhs.dims[1]:
            state = Qobj(unstack_columns(data),
                         **self._state_metadata, copy=False)
        else:
            state = Qobj(data, **self._state_metadata, copy=copy)

        return state

    def _initialize_stats(self):
        stats = super()._initialize_stats()
        stats.update({
            "method": self.options["method"],
            "solver": "Master Equation Evolution",
            "num_collapse": self._num_collapse,
        })
        return stats

    def _argument(self, args):
        self._integrator.arguments(args)
        self.rhs.arguments(args)
        for c_op in self._c_ops:
            c_op.arguments(args)
        for n_op in self._n_ops:
            n_op.arguments(args)

    def _run_one_traj(self, seed, state, tlist, e_ops):
        """
        Run one trajectory and return the result.
        """
        # The integrators is reused, but non-reentrant. They are are fine for
        # multiprocessing, but will fail with multithreading.
        # If a thread base parallel map is created, eahc trajectory should use
        # a copy of the integrator.
        result = Result(e_ops, {**self.options, "normalize_output": False})
        generator = self._get_generator(seed)
        self._integrator.set_state(tlist[0], state, generator)
        result.add(tlist[0], self._restore_state(state, copy=False))
        for t in tlist[1:]:
            t, state = self._integrator.integrate(t, copy=False)
            result.add(t, self._restore_state(state, copy=False))
        result.collapse = self._integrator.collapses
        return seed, result

    def _get_integrator(self):
        _time_start = time()
        method = self.options["method"]
        if method in self.avail_integrators():
            integrator = self.avail_integrators()[method]
        elif issubclass(method, Integrator):
            integrator = method
        else:
            raise ValueError("Integrator method not supported.")
        integrator_instance = integrator(self.rhs, self.options)
        mc_integrator = MCIntegrator(
            integrator_instance, self._c_ops, self._n_ops, self.options
        )
        self._init_integrator_time = time() - _time_start
        return mc_integrator

    @property
    def options(self):
        """
        Options for monte carlo solver:

        store_final_state: bool, default=False
            Whether or not to store the final state of the evolution in the
            result class.

        store_states: bool, default=None
            Whether or not to store the state vectors or density matrices.
            On `None` the states will be saved if no expectation operators are
            given.

        progress_bar: str {'text', 'enhanced', 'tqdm', ''}, default="text"
            How to present the solver progress.
            'tqdm' uses the python module of the same name and raise an error if
            not installed. Empty string or False will disable the bar.

        progress_kwargs: dict, default={"chunk_size":10}
            Arguments to pass to the progress_bar. Qutip's bars use
            ``chunk_size``.

        keep_runs_results: bool
          Whether to store results from all trajectories or just store the
          averages.

        method: str, default="adams"
            Which ODE integrator methods are supported.

        map: str {"serial", "parallel", "loky"}
            How to run the trajectories. "parallel" uses concurent module to
            run in parallel while "loky" use the module of the same name to do
            so.

        job_timeout: None, int
            Maximum time to compute one trajectory.

        num_cpus: None, int
            Number of cpus to use when running in parallel. ``None`` detect the
            number of available cpus.

        bitgenerator: {None, "MT19937", "PCG64", "PCG64DXSM", ...}
            Which of numpy.random's bitgenerator to use. With ``None``, your
            numpy version's default is used.

        mc_corr_eps: float
            Small number used to detect non-physical collapse caused by
            numerical imprecision.

        norm_t_tol: float
            Tolerance in time used when finding the collapse.

        norm_tol: float
            Tolerance in norm used when finding the collapse.

        norm_steps: int
            Maximum number of tries to find the collapse.
        """
        return self._options

    @options.setter
    def options(self, new_options):
        MultiTrajSolver.options.fset(self, new_options)

    @classmethod
    def avail_integrators(cls):
        if cls is Solver:
            return cls._avail_integrators.copy()
        return {
            **Solver.avail_integrators(),
            **cls._avail_integrators,
        }



#cython: language_level=3
from qutip.core.data cimport Data
from qutip.core.cy.qobjevo cimport QobjEvo

cpdef enum Status:
    AT_FRONT = 2
    INTERPOLATED = 1
    NORMAL = 0
    TOO_MUCH_WORK = -1
    DT_UNDERFLOW = -2
    OUTSIDE_RANGE = -3
    NOT_INITIATED = -4

cdef class Explicit_RungeKutta:
    cdef QobjEvo qevo

    # Ode state data, set in set_initial_value
    cdef list k
    cdef Data _y_temp, _y, _y_prev, _y_front
    cdef double _norm_front, _norm_prev, _dt_safe, _dt_int
    cdef double _t, _t_prev, _t_front
    cdef Status _status
    cdef dict status_messages

    # options: set in init
    cdef readonly double rtol, atol, first_step, min_step, max_step
    cdef readonly int max_numsteps
    cdef readonly bint interpolate
    cdef readonly object method

    # Runge Kutta tableau and info
    cdef int rk_step, rk_extra_step, order, denseout_order
    cdef bint adaptative_step, can_interpolate
    cdef object b_factor_np
    cdef double [:] b
    cdef double [:] b_factor
    cdef double [:] c
    cdef double [:] e
    cdef double [:, ::1] a
    cdef double [:, ::1] bi

    cpdef integrate(Explicit_RungeKutta self, double t, bint step=*)

    cpdef void set_initial_value(self, Data y0, double t)

    cdef int _step_in_err(self, double t, int max_step)

    cdef double _compute_step(self, double dt)

    cdef double _error(self, Data y_new, double dt)

    cdef void _prep_dense_out(self)

    cdef Data _interpolate_step(self, double t, Data out)

    cdef inline Data _accumulate(self, Data target, double[:] factors,
                                 double dt, int size)

    cdef double _estimate_first_step(self, double t, Data y0)

    cdef double _get_timestep(self, double t)

    cdef void _recompute_safe_step(self, double err, double dt)



#cython: language_level=3

import functools
import inspect
import itertools
import warnings

from .convert import to as _to

cimport cython
from libc cimport math
from libcpp cimport bool

__all__ = ['Dispatcher']

cdef class _bind:
    """
    Cythonised implementation of inspect.Signature.bind, supporting faster
    binding and handling of default arguments.  On construction, the signature
    of the base function is extracted, and all parameters are parsed into
    positional or keyword slots, using positional wherever possible, and their
    defaults are stored.
    """
    # Instance of inpsect.Signature representing the function.
    cdef object signature
    cdef object inputs
    # Mapping of (str: int), where str is the name of any argument which may be
    # specified by a keyword, and int is its location in `self._pos` if
    # available, or -1 if it is keyword-only.
    cdef dict _locations
    # Default values (or inspect.Parameter.empty) for every parameter which
    # may be passed as a positional argument.
    cdef list _pos
    # For each dispatcher input which can be given as a positional argument,
    # these two lists hold the index of the input in the given tuple of input
    # names and the corresponding index into the positional arguments after
    # binding is complete.  {_pos_inputs_pos[n]: _pos_inputs_input[n]} is
    # effectively the same mapping as _kw_inputs (but for positional
    # arguments), but we use two lists rather than a dictionary for speed.
    cdef list _pos_inputs_input, _pos_inputs_pos
    # Default values (or inspect.Parameter.empty) for every parameter which
    # _must_ be passed as a keyword argument.
    cdef dict _kw
    # Mapping of (name, index into the input tuple) for each input which must
    # be specified as a keyword argument
    cdef list _kw_inputs
    # Names of the keyword arguments which have default values.
    cdef set _default_kw_names
    # Respectively, numbers of positional parameters, keyword-only parameters
    # and dispatcher inputs.
    cdef Py_ssize_t n_args, n_kwargs, n_inputs
    # Respectively, numbers of inputs which are positional and keyword-only.
    cdef Py_ssize_t _n_pos_inputs, _n_kw_inputs
    # Respectively, numbers of positional arguments without a default, and
    # number of keyword-only arguments which _have_ a default.
    cdef Py_ssize_t _n_pos_no_default, _n_kw_default


    def __init__(self, signature, tuple inputs):
        for arg in inputs:
            if arg not in signature.parameters:
                raise AttributeError("No argument matches '{}'.".format(arg))
        self.signature = signature
        self.inputs = inputs
        self._locations = {}
        self._pos = []
        self._pos_inputs_input = []
        self._pos_inputs_pos = []
        self._kw = {}
        self._kw_inputs = []
        self._default_kw_names = set()
        self._n_pos_no_default = 0
        # signature.parameters is ordered for all Python versions.
        for i, (name, parameter) in enumerate(signature.parameters.items()):
            kind = parameter.kind
            if (kind == parameter.VAR_POSITIONAL
                    or kind == parameter.VAR_KEYWORD):
                raise TypeError("Cannot dispatch with *args or **kwargs.")
            if kind == parameter.KEYWORD_ONLY:
                self._kw[name] = parameter.default
                self._locations[name] = -1
                if parameter.default is not parameter.empty:
                    self._default_kw_names.add(name)
                if name in inputs:
                    self._kw_inputs.append((name, inputs.index(name)))
            else:
                self._pos.append(parameter.default)
                if kind != parameter.POSITIONAL_ONLY:
                    self._locations[name] = i
                if parameter.default is parameter.empty:
                    self._n_pos_no_default += 1
                if name in inputs:
                    # Effectively a mapping.
                    self._pos_inputs_input.append(inputs.index(name))
                    self._pos_inputs_pos.append(len(self._pos) - 1)
        self.n_args = len(self._pos)
        self.n_kwargs = len(self._kw)
        self.n_inputs = len(inputs)
        self._n_pos_inputs = len(self._pos_inputs_input)
        self._n_kw_inputs = len(self._kw_inputs)
        self._n_kw_default = len(self._default_kw_names)


    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef tuple bind(self, tuple args, dict kwargs):
        """
        Cython reimplementation of inspect.Signature.bind for binding the
        collected `*args` and `**kwargs` of a generic call to a specific
        signature, checking that everything matches.

        The output is a parsed tuple (args, kwargs), where `args` is a list and
        `kwargs` is a dict, however all arguments which _can_ be passed
        positionally will be moved from `kwargs` into `args` for the output.
        The resultant `args` and `kwargs` can be unpacked into the underlying
        function call safely.  Default values are not filled in by this method.

        This is necessary to allow a generic `Dispatcher` class to work with
        all type signatures.  The result of this function can be fed to
        `_bind.convert_types` and `_bind.dispatch_types`.
        """
        cdef Py_ssize_t location, got_pos=0, got_kw=0
        cdef Py_ssize_t n_passed_args = len(args)
        if n_passed_args > self.n_args:
            raise TypeError(
                "Too many arguments: expected at most {} but received {}."
                .format(n_passed_args, self.n_args)
            )
        # _pos and _kw are filled with their defaults at initialisation, and
        # _pos is already the correct length (no-default parameters have
        # sentinel values).
        cdef list out_pos = self._pos.copy()
        cdef dict out_kw = self._kw.copy()
        got_pos = self.n_args - self._n_pos_no_default
        got_kw = self._n_kw_default
        # Positional arguments unambiguously fill the first positional slots.
        for location in range(n_passed_args):
            out_pos[location] = args[location]
        # How many arguments that we didn't have defaults for did we just get?
        # Positionals with a default are always after those without one.
        if n_passed_args > self._n_pos_no_default:
            got_pos += self._n_pos_no_default
        else:
            got_pos += n_passed_args
        # Everything else has been passed by keyword, but it may be allowed to
        # be passed positionally, which is the case we prefer.  dict.items() is
        # (relatively) expensive, so we use a boolean test for the fast path.
        cdef str kw
        cdef object arg
        if kwargs:
            for kw, arg in kwargs.items():
                try:
                    location = self._locations[kw]
                except KeyError:
                    raise TypeError("Unknown argument '{}'.".format(kw)) from None
                # _locations[kw] = -1 if kw is keyword-only, otherwise the
                # corresponding positional location.
                if location >= 0:
                    if location < n_passed_args:
                        raise TypeError("Multiple values for '{}'".format(kw))
                    out_pos[location] = arg
                    if location < self._n_pos_no_default:
                        got_pos += 1
                else:
                    out_kw[kw] = arg
                    if kw not in self._default_kw_names:
                        got_kw += 1
        if got_pos < self.n_args:
            raise TypeError("Too few positional arguments passed.")
        if got_kw < self.n_kwargs:
            raise TypeError("Not all keyword arguments were filled.")
        return out_pos, out_kw

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef list dispatch_types(self, list args, dict kwargs):
        """
        Get a list of the types which the dispatch should operate over, given
        the output of `_bind.bind`.  The return value is a list of the _input_
        types to be dispatched on, in order that they were specified at
        `Dispatcher` creation.  The output type is not returned as part of this
        list, because there is no way for `_bind` to know it.
        """
        cdef list dispatch = [None] * self.n_inputs
        cdef str kw
        cdef Py_ssize_t location, i
        for location in range(self._n_pos_inputs):
            dispatch[self._pos_inputs_input[location]]\
                = type(args[self._pos_inputs_pos[location]])
        for i in range(self._n_kw_inputs):
            kw, location = self._kw_inputs[i]
            dispatch[location] = type(kwargs[kw])
        return dispatch

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef tuple convert_types(self, list args, dict kwargs, tuple converters):
        """
        Apply the type converters `converters` to the parsed arguments `args`
        and `kwargs`.

        If there are `n` inputs which are dispatched on, `converters` should be
        a tuple whose first `n` elements are converters (such as those obtained
        from `data.to[to_type, from_type]`) to the desired types.

        `args` and `kwargs` should be the output of `_bind.bind`; the function
        will likely fail if called on unparsed arguments.
        """
        cdef str kw
        cdef Py_ssize_t location, i
        for location in range(self._n_pos_inputs):
            args[location] = converters[location](args[location])
        for i in range(self._n_kw_inputs):
            kw, location = self._kw_inputs[i]
            kwargs[kw] = converters[location](kwargs[kw])
        return args, kwargs


cdef double _conversion_weight(tuple froms, tuple tos, dict weight_map, bint out) except -1:
    """
    Find the total weight of conversion if the types in `froms` are converted
    element-wise to the types in `tos`.  `weight_map` is a mapping of
    `(to_type, from_type): real`; it should almost certainly be
    `data.to.weight`.
    """
    cdef double weight = 0.0
    cdef Py_ssize_t i, n=len(froms)
    if len(tos) != n:
        raise ValueError(
            "number of arguments not equal: " + str(n) + " and " + str(len(tos))
        )
    if out:
        n = n - 1
        weight += weight_map[froms[n], tos[n]]
    for i in range(n):
        weight += weight_map[tos[i], froms[i]]
    return weight


cdef class _constructed_specialisation:
    """
    Callable object providing the specialisation of a data-layer operation for
    a particular set of types (`self.types`).  This may or may not involve
    conversion of the input types and the output to match a known
    specialisation; if it has no conversions, `self.direct` will be `True`,
    otherwise it will be `False`.

    See `self.__signature__` or `self.__text_signature__` for the call
    signature of this object.
    """
    cdef _bind _parameters
    cdef bint _output
    cdef object _call
    cdef Py_ssize_t _n_dispatch
    cdef readonly tuple types
    cdef tuple _converters
    cdef str _short_name
    cdef public str __doc__
    cdef public str __name__
    cdef public str __module__
    cdef public object __signature__
    cdef readonly str __text_signature__
    cdef readonly bint direct

    def __init__(self, base, Dispatcher dispatcher, types, converters, out,
                 direct):
        self.__doc__ = inspect.getdoc(dispatcher)
        self._short_name = dispatcher.__name__
        self.__name__ = (
            self._short_name
            + "_"
            + "_".join([x.__name__ for x in types])
        )
        self.__module__ = dispatcher.__module__
        self.__signature__ = dispatcher.__signature__
        self.__text_signature__ = dispatcher.__text_signature__
        self._parameters = dispatcher._parameters
        self._output = out
        self._call = base
        self._n_dispatch = len(types)
        self.types = types
        self.direct = direct
        self._converters = converters

    cdef object prebound(self, list args, dict kwargs):
        """
        Call this specialisation with pre-parsed arguments and keyword
        arguments.  `args` and `kwargs` must be the output of the relevant
        `_bind.bind` method for this function.
        """
        self._parameters.convert_types(args, kwargs, self._converters)
        out = self._call(*args, **kwargs)
        if self._output:
            out = self._converters[self._n_dispatch - 1](out)
        return out

    def __call__(self, *args, **kwargs):
        cdef list args_
        cdef dict kwargs_
        args_, kwargs_ = self._parameters.bind(args, kwargs)
        return self.prebound(args_, kwargs_)

    def __repr__(self):
        if len(self.types) == 1:
            spec = self.types[0].__name__
        else:
            spec = "(" + ", ".join(x.__name__ for x in self.types) + ")"
        direct = ("" if self.direct else "in") + "direct"
        return "".join([
            "<", direct, " specialisation ", spec, " of ", self._short_name, ">"
        ])


cdef class Dispatcher:
    """
    Dispatcher for a data-layer operation.  This object can be called with the
    signature shown in `self.__signature__` or `self.__text_signature__`, where
    the arguments listed in `self.inputs` can be any data-layer types (i.e.
    ones that have valid conversions in `data.to`).

    You can define additional specialisations for this dispatcher by calling
    its `add_specialisations` method.  New data types must be added to
    `data.to` before they can be added as specialisations to a dispatcher.

    You can get a callable object representing a single set of dispatcher types
    by using the key-lookup syntax
        Dispatcher[type1, type2, ...]
    where `type1`, `type2`, etc are the dispatched arguments (with the output
    type on the end, if this is a dispatcher over the output type.
    """
    cdef _bind _parameters
    cdef readonly dict _specialisations
    cdef Py_ssize_t _n_dispatch
    cdef readonly dict _lookup
    cdef set _dtypes
    cdef bint _pass_on_dtype
    cdef readonly tuple inputs
    cdef readonly bint output
    cdef public str __doc__
    cdef public str __name__
    cdef public str __module__
    cdef public object __signature__
    cdef readonly str __text_signature__

    def __init__(self, signature_source, inputs, bint out=False,
                 str name=None, str module=None):
        """
        Create a new data layer dispatching operator.

        Parameters
        ----------
        signature_source : callable or inspect.Signature
            An object from which the call signature of operation can be
            determined.  You can pass any callable defined in Python space, and
            the signature will be extracted.  Note that the callable will not
            be added as a specialisation by this; you will still have to call
            `add_specialisations`.

            If you cannot provide a callable with an extractable signature
            (e.g. Cython extension methods), you can instead directly provide
            an instance of `inspect.Signature`, which will be used instead.

        inputs : iterable of str
            The parameters which should be dispatched over.  These can be
            positional or keyword arguments, but must feature in the signature
            provided.

        out : bool, optional (False)
            Whether to dispatch on the output of the function.  Defaults to
            `False`.

        name : str, optional
            If given, the `__name__` parameter of the dispatcher is set to
            this.  If not given and `signature_source` is _not_ an instance of
            `inspect.Signature`, then we will attempt to read `__name__` from
            there instead.

        module : str, optional
            If given, the `__module__` parameter of the dispatcher is set to
            this.  If not given and `signature_source` is _not_ an instance of
            `inspect.Signature`, then we will attempt to read `__module__` from
            there instead.
        """
        if isinstance(inputs, str):
            inputs = (inputs,)
        inputs = tuple(inputs)
        if inputs == () and out is False:
            warnings.warn(
                "No parameters to dispatch on."
                " Maybe you meant to specify 'inputs' or 'out'?"
            )
        self.inputs = inputs
        if isinstance(signature_source, inspect.Signature):
            self.__signature__ = signature_source
        else:
            self.__signature__ = inspect.signature(signature_source)
        self._parameters = _bind(self.__signature__, inputs)
        if name is not None:
            self.__name__ = name
        elif not isinstance(signature_source, inspect.Signature):
            self.__name__ = signature_source.__name__
        else:
            self.__name__ = 'dispatcher'
        if module is not None:
            self.__module__ = module
        elif not isinstance(signature_source, inspect.Signature):
            self.__module__ = signature_source.__module__
        self.__text_signature__ = self.__name__ + str(self.__signature__)
        if not isinstance(signature_source, inspect.Signature):
            self.__doc__ = inspect.getdoc(signature_source)
        self.output = out
        self._specialisations = {}
        self._lookup = {}
        self._n_dispatch = self._parameters.n_inputs + self.output
        self._pass_on_dtype = 'dtype' in self.__signature__.parameters
        # Add ourselves to the list of dispatchers to be updated.
        _to.dispatchers.append(self)

    def add_specialisations(self, specialisations, _defer=False):
        """
        Add specialisations for particular combinations of data types to this
        operation.  The data types must already be known in `data.to` before
        you try to provide them here.  All data types defined in `data.to` will
        automatically work with this dispatcher, but will involve inefficient
        conversions to and from other types unless you define a closer
        specialisation using this method.

        The lookup table will automatically be rebuilt after this method is
        called.  Specialisations defined more than once will use the most
        recent version; you can use this to override currently known
        specialisations if desired.

        Parameters
        ----------
        specialisations : iterable of tuples
            An iterable where each element specifies a new specialisation for
            this operation.  Each element of the iterable should be a tuple,
            whose items are the types (instances of `type`) which this
            specialisation takes in each of the slots defined by
            `Dispatcher.inputs`, and the output type if this is a dispatcher
            over output types.  The last element should be the callable itself.

            The callable must have exactly the same signature as
            `Dispatcher.__signature__`; it is not enough that it takes all the
            same keyword arguments, but they must come in the same order as
            well (this is a speed optimisation for the dispatching operation).

            For example, if this is a dispatcher with the signature
                add(left, right, scale=1)
            which also dispatches over its output, and we have specialisations
                add_1(left: CSR, right: Dense, scale=1) -> Dense
                add_2(left: Dense, right: CSC, scale=1) -> CSR
            then to add this, `specialisations` should look like
                [
                    (CSR, Dense, Dense, add_1),
                    (Dense, CSC, CSR, add_2),
                ]
            Type annotations present in the specialisation objects are ignored.

        _defer : bool, optional (False)
            Only intended for internal library use during initialisation. If
            `True`, then the input types are not checked, and the full lookup
            table is not built until a manual call to
            `Dispatcher.rebuild_lookup()` is made.  If you are getting errors,
            remember that you should add the data type conversions to `data.to`
            before you try to add specialisations.
        """
        for arg in specialisations:
            arg = tuple(arg)
            if len(arg) != self._n_dispatch + 1:
                raise ValueError(
                    "specialisation " + str(arg)
                    + " has wrong number of parameters: needed types for "
                    + str(self._parameters.inputs)
                    + (", an output type" if self.output else "")
                    + " and a callable"
                )
            for i in range(self._n_dispatch):
                if (not _defer) and arg[i] not in _to.dtypes:
                    raise ValueError(str(arg[i]) + " is not a known data type")
            if not callable(arg[self._n_dispatch]):
                raise TypeError(str(arg[-1]) + " is not callable")
            self._specialisations[arg[:-1]] = arg[-1]
        if not _defer:
            self.rebuild_lookup()

    def rebuild_lookup(self):
        """
        Manually trigger a rebuild of the lookup table for this dispatcher.
        This is called automatically when new data types are added to
        `data.to`, or when specialisations are added to this object with
        `Dispatcher.add_specialisations`.

        You most likely do not need to call this function yourself.
        """
        cdef double weight, cur
        cdef tuple types, out_types
        cdef object function
        if not self._specialisations:
            return
        self._dtypes = _to.dtypes.copy()
        # The complexity of building the table here is very poor, but it's a
        # cost we pay very infrequently, and until it's proved to be a
        # bottle-neck in real code, we stick with the simple algorithm.
        for in_types in itertools.product(self._dtypes, repeat=self._n_dispatch):
            weight = math.INFINITY
            types = None
            function = None
            for out_types, out_function in self._specialisations.items():
                cur = _conversion_weight(in_types, out_types, _to.weight,
                                         out=self.output)
                if cur < weight:
                    weight = cur
                    types = out_types
                    function = out_function
            if self.output:
                converters = tuple(
                    [_to[pair] for pair in zip(types[:-1], in_types[:-1])]
                    + [_to[in_types[-1], types[-1]]]
                )
            else:
                converters = tuple(_to[pair] for pair in zip(types, in_types))
            self._lookup[in_types] =\
                _constructed_specialisation(function, self, in_types,
                                            converters, self.output,
                                            weight == 0)
        # Now build the lookup table in the case that we dispatch on the output
        # type as well, but the user has called us without specifying it.
        # TODO: option to control default output type choice if unspecified?
        if self.output:
            for in_types in itertools.product(self._dtypes, repeat=self._n_dispatch-1):
                weight = math.INFINITY
                types = None
                function = None
                for out_types, out_function in self._specialisations.items():
                    cur = _conversion_weight(in_types, out_types[:-1],
                                             _to.weight, out=False)
                    if cur < weight:
                        weight = cur
                        types = out_types
                        function = out_function
                converters = tuple(_to[pair] for pair in zip(types, in_types))
                self._lookup[in_types] =\
                    _constructed_specialisation(function, self,
                                                in_types + (types[-1],),
                                                converters, False,
                                                weight == 0)

    def __getitem__(self, types):
        """
        Get the particular specialisation for the given types.  The output is a
        callable object which requires that the dispatched arguments match
        those specified in `types`.
        """
        if type(types) is not tuple:
            types = (types,)
        types = tuple(_to.parse(arg) for arg in types)
        try:
            return self._lookup[types]
        except KeyError:
            raise TypeError("specialisation not known for types: " + str(types)) from None

    def __repr__(self):
        return "<dispatcher: " + self.__text_signature__ + ">"

    def __call__(self, *args, dtype=None, **kwargs):
        cdef list args_, dispatch
        cdef dict kwargs_
        if self._pass_on_dtype:
            kwargs['dtype'] = dtype
        if not (self._pass_on_dtype or self.output) and dtype is not None:
            raise TypeError("unknown argument 'dtype'")
        args_, kwargs_ = self._parameters.bind(args, kwargs)
        dispatch = self._parameters.dispatch_types(args_, kwargs_)
        if self.output and dtype is not None:
            dtype = _to.parse(dtype)
            dispatch.append(dtype)
        cdef _constructed_specialisation function
        try:
            function = self._lookup[tuple(dispatch)]
        except KeyError:
            raise TypeError("unknown types to dispatch on: " + str(dispatch)) from None
        return function.prebound(args_, kwargs_)



"""
This module is a collection of random state and operator generators.
"""

__all__ = [
    'rand_herm',
    'rand_unitary',
    'rand_dm',
    'rand_stochastic',
    'rand_ket',
    'rand_kraus_map',
    'rand_super',
    "rand_super_bcsz",
]

import numbers

import numpy as np
from numpy.random import Generator, SeedSequence, default_rng
import scipy.linalg
import scipy.sparse as sp

from . import (Qobj, create, destroy, jmat, basis,
               to_super, to_choi, to_chi, to_kraus, to_stinespring)
from .core import data as _data
from .core.dimensions import flatten


_RAND = default_rng()
_UNITS = np.array([1, 1j])


def _implicit_tensor_dimensions(dimensions, superoper=False):
    """
    Total flattened size and operator dimensions for operator creation routines
    that automatically perform tensor products.

    Parameters
    ----------
    dimensions : (int) or (list of int) or (list of list of int)
        First dimension of an operator which can create an implicit tensor
        product.  If the type is `int`, it is promoted first to `[dimensions]`.
        From there, it should be one of the two-elements `dims` parameter of a
        `qutip.Qobj` representing an `oper` or `super`, with possible tensor
        products.

    Returns
    -------
    size : int
        Dimension of backing matrix required to represent operator.
    dimensions : list
        Dimension list in the form required by ``Qobj`` creation.
    """
    if not isinstance(dimensions, list):
        dimensions = [dimensions]
    flat = flatten(dimensions)
    if not all(isinstance(x, numbers.Integral) and x >= 0 for x in flat):
        raise ValueError("All dimensions must be integers >= 0")
    N = np.prod(flat)
    if superoper:
        if isinstance(dimensions[0], numbers.Integral):
            dimensions = [dimensions, dimensions]
        else:
            N = int(N**0.5)
    return N, [dimensions, dimensions]


def _get_generator(seed):
    """
    Obtain a random generator from a seed or generator.

    Parameters
    ----------
    seed: int, SeedSequence, Generator, NoneType
        Seed to create the generator. If it's already a generator return it.
        When ``None`` is suplied, a default generator is provided.
    """
    if seed is None:
        gen = _RAND
    elif isinstance(seed, Generator):
        gen = seed
    else:
        gen = default_rng(seed)
    return gen


def _randnz(shape, generator, norm=np.sqrt(0.5)):
    """
    Returns an array of standard normal complex random variates.
    The Ginibre ensemble corresponds to setting ``norm = 1`` [Mis12]_.

    Parameters
    ----------
    shape : tuple
        Shape of the returned array of random variates.

    generator : Generator
        Random number generator.

    norm : float
        Scale of the returned random variates, or 'ginibre' to draw
        from the Ginibre ensemble.
    """
    # This function is intended for internal use.
    if norm == 'ginibre':
        norm = 1
    return np.sum(generator.normal(size=(shape + (2,))) * _UNITS, axis=-1) * norm


def _rand_jacobi_rotation(A, generator):
    """Random Jacobi rotation of a matrix.

    Parameters
    ----------
    A : qutip.data.Data
        Matrix to rotate as a data layer object.

    generator : numpy.random.generator
        Random number generator.

    Returns
    -------
    qutip.data.Data
        Rotated sparse matrix.
    """
    if A.shape[0] != A.shape[1]:
        raise ValueError('Input matrix must be square.')
    n = A.shape[0]
    angle = 2 * generator.random() * np.pi
    a = np.sqrt(0.5) * np.exp(-1j * angle)
    b = np.conj(a)
    i = generator.integers(n)
    j = i
    while i == j:
        j = generator.integers(n)
    data = np.hstack((np.array([a, -b, a, b], dtype=complex),
                      np.ones(n - 2, dtype=complex)))
    diag = np.delete(np.arange(n), [i, j])
    rows = np.hstack(([i, i, j, j], diag))
    cols = np.hstack(([i, j, i, j], diag))
    R = sp.coo_matrix((data, (rows, cols)), shape=(n, n), dtype=complex)
    R = _data.create(R.tocsr())
    return _data.matmul(_data.matmul(R, A), R.adjoint())


def _get_block_sizes(N, density, generator):
    """
    Obtain a list of matrix block sizes in such a way that an NxN matrix
    composed of full matrices of these sizes along the diagonal would be of
    desired density.
    """
    if density <= 0:
        return [1] * N
    elif density >= 1:
        return [N]
    max_block_size = int((N**2 * density)**0.5)
    min_block_size = 1
    if density >= 0.5:
        min_block_size = int(N / 2 * (1 + (2 * density - 1)**0.5))

    if min_block_size >= max_block_size:
        M = min(min_block_size, max_block_size)
    else:
        M = generator.integers(min_block_size, max_block_size) + 1
    M += generator.integers(-1, 2)
    M = min(max(M, 1), N)

    other_blocks = []
    if M < N:
        other_N = N - M
        other_density = (density * N**2 - M**2) / other_N**2
        other_blocks = _get_block_sizes(other_N, other_density, generator)

    return [M] + other_blocks


def _merge_shuffle_blocks(blocks, generator):
    """
    For a list of block, merge them in one matrix along the diagonal and
    shuffle the rows and columns.
    """
    N = sum(block.shape[0] for block in blocks)
    D = sum(block.shape[0]**2 for block in blocks) / N**2
    idx = generator.permutation(N)
    end = 0
    if D < 0.1:
        # TODO: coo_array from sicpy 1.8 would allow to simplify this.
        row = []
        col = []
        data = []
        for block in blocks:
            start = end
            end = end + block.shape[0]
            brow, bcol = np.meshgrid(idx[start:end], idx[start:end])
            row.append(brow.ravel())
            col.append(bcol.ravel())
            data.append(block.ravel())
        data = np.hstack(data)
        row = np.hstack(row)
        col = np.hstack(col)
        matrix = sp.coo_matrix((data, (row, col)), shape=(N, N))
    else:
        matrix = np.zeros((N,N), dtype=complex)
        for block in blocks:
            start = end
            end = end + block.shape[0]
            brow, bcol = np.meshgrid(idx[start:end], idx[start:end])
            matrix[brow, bcol] = block
    return _data.create(matrix, copy=False)


def rand_herm(dimensions, density=0.30, distribution="fill", *,
              eigenvalues=(), seed=None, dtype=_data.CSR):
    """Creates a random sparse Hermitian quantum object.

    Parameters
    ----------
    dimensions : (int) or (list of int) or (list of list of int)
        Dimension of Hilbert space. If provided as a list of ints, then the
        dimension is the product over this list, but the ``dims`` property of
        the new Qobj are set to this list.  This can produce either `oper` or
        `super` depending on the passed `dimensions`.

    density : float, [0.30]
        Density between [0,1] of output Hermitian operator.

    distribution : str {"fill", "pos_def", "eigen"}
        Method used to obtain the density matrices.

        - "fill" : Uses :math:`H=0.5*(X+X^{+})` where :math:`X` is a randomly
          generated quantum operator with elements uniformly distributed
          between ``[-1, 1] + [-1j, 1j]``.
        - "eigen" : A density matrix with the given ``eigenvalues``. It uses
          random complex Jacobi rotations to shuffle the operator.
        - "pos_def" : Return a positive semi-definite matrix by diagonal
          dominance.

    eigenvalues : array_like, optional
        Eigenvalues of the output Hermitian matrix. The len must match the
        shape of the matrix.

    seed : int, SeedSequence, Generator, optional
        Seed to create the random number generator or a pre prepared
        generator. When none is suplied, a default generator is used.

    dtype : type or str
        Storage representation. Any data-layer known to `qutip.data.to` is
        accepted.

    Returns
    -------
    oper : :class:`qobj`
        Hermitian quantum operator.

    Notes
    -----
    If given a list of eigenvalues the object is created using complex Jacobi
    rotations.  While this method is fast for small matrices, it should not be
    repeatedly used for generating matrices larger than ~1000x1000.
    """
    N, dims = _implicit_tensor_dimensions(dimensions)
    generator = _get_generator(seed)
    if distribution not in ["eigen", "fill", "pos_def"]:
        raise ValueError("distribution must be one of {'eigen', 'fill', "
                         "'pos_def'}")

    if distribution == "eigen":
        if N != len(eigenvalues):
            raise ValueError("The number of eigenvalues does not match the "
                             "desired shape.")
        out = _data.diag[_data.CSR](eigenvalues, 0)
        nvals = max([N**2 * density, 1])
        out = _rand_jacobi_rotation(out, generator)
        while _data.csr.nnz(out) < 0.95 * nvals:
            out = _rand_jacobi_rotation(out, generator)
        out = Qobj(out, type='oper', dims=dims, isherm=True, copy=False)

    else:
        pos_def = distribution == "pos_def"
        if density < 0.5:
            M = _rand_herm_sparse(N, density, pos_def, generator)
        else:
            M = _rand_herm_dense(N, density, pos_def, generator)

        out = Qobj(M, type='oper', dims=dims, isherm=True, copy=False)

    return out.to(dtype)


def _rand_herm_sparse(N, density, pos_def, generator):
    target = (1 - (1 - density)**0.5)
    num_elems = (N**2 - 0.666 * N) * target + 0.666 * N * density
    num_elems = max([num_elems, 1])
    num_elems = int(num_elems)
    data = (2 * generator.random(num_elems) - 1) + \
           (2 * generator.random(num_elems) - 1) * 1j
    row_idx, col_idx = zip(*[
        divmod(index, N)
        for index in generator.choice(N*N, num_elems, replace=False)
    ])
    M = sp.coo_matrix((data, (row_idx, col_idx)),
                      dtype=complex, shape=(N, N))
    M = 0.5 * (M + M.conj().transpose())
    M.sort_indices()
    rand_mat = _data.create(M)
    if pos_def:
        rand_mat = _data.add(
            rand_mat,
            _data.diag(np.ones(N, dtype=complex)),
            np.sqrt(2) * N + 1
        )
    return rand_mat


def _rand_herm_dense(N, density, pos_def, generator):
    M = (
        (2*generator.random((N, N)) - 1)
        + 1j*(2*generator.random((N, N)) - 1)
    )
    M = 0.5 * (M + M.conj().transpose())
    target = 1 - density**0.5
    num_remove = N * (N - 0.666) * target + 0.666 * N * (1 - density)
    num_remove = max([num_remove, 1])
    num_remove = int(num_remove)
    for index in generator.choice(N*N, num_remove, replace=False):
        row, col = divmod(index, N)
        M[col, row] = 0
        M[row, col] = 0
    if pos_def:
        np.fill_diagonal(M, np.abs(M.diagonal()) + np.sqrt(2) * N )
    return _data.create(M)


def rand_unitary(dimensions, density=1, distribution="haar", *,
                 seed=None, dtype=_data.Dense):
    r"""Creates a random sparse unitary quantum object.

    Parameters
    ----------
    dimensions : (int) or (list of int) or (list of list of int)
        Dimension of Hilbert space. If provided as a list of ints, then the
        dimension is the product over this list, but the ``dims`` property of
        the new Qobj are set to this list.  This can produce either `oper` or
        `super` depending on the passed `dimensions`.

    density : float, [1]
        Density between [0,1] of output unitary operator.

    distribution : ["haar", "exp"]
        Method used to obtain the unitary matrices.

        - haar : Haar random unitary matrix using the algorithm of [Mez07]_.
        - exp : Uses :math:`\exp(-iH)`, where H is a randomly generated
          Hermitian operator.

    seed : int, SeedSequence, Generator, optional
        Seed to create the random number generator or a pre prepared
        generator. When none is suplied, a default generator is used.

    dtype : type or str
        Storage representation. Any data-layer known to `qutip.data.to` is
        accepted.

    Returns
    -------
    oper : qobj
        Unitary quantum operator.
    """
    N, dims = _implicit_tensor_dimensions(dimensions)
    if distribution not in ["haar", "exp"]:
        raise ValueError("distribution must be one of {'haar', 'exp'}")
    generator = _get_generator(seed)
    block_sizes = _get_block_sizes(N, density, generator)

    blocks = []
    for block in block_sizes:
        if distribution == "haar":
            mat = _rand_unitary_haar(block, generator)
        elif distribution == "exp":
            mat = scipy.linalg.expm(
                -1.0j *
                _rand_herm_dense(block, 1, False, generator).as_ndarray()
            )
        blocks.append(mat)

    merged = _merge_shuffle_blocks(blocks, generator)

    mat = Qobj(merged, type='oper', dims=dims, isunitary=True, copy=False)
    return mat.to(dtype)


def _rand_unitary_haar(N, generator):
    """
    Returns a Haar random unitary matrix of dimension
    ``dim``, using the algorithm of [Mez07]_.

    Parameters
    ----------
    N : int
        Dimension of the unitary to be returned.

    generator : numpy.random.generator
        Random number generator.

    Returns
    -------
    U : qutip.core.data.Dense
        Unitary of shape ``(N, N)`` drawn from the Haar measure.
    """
    # Mez01 STEP 1: Generate an N × N matrix Z of complex standard
    #               normal random variates.
    Z = _randnz((N, N), generator)

    # Mez01 STEP 2: Find a QR decomposition Z = Q · R.
    Q, R = scipy.linalg.qr(Z)

    # Mez01 STEP 3: Create a diagonal matrix Lambda by rescaling
    #               the diagonal elements of R.
    Lambda = np.diag(R).copy()
    Lambda /= np.abs(Lambda)

    # Mez01 STEP 4: Note that R' := Λ¯¹ · R has real and
    #               strictly positive elements, such that
    #               Q' = Q · Λ is Haar distributed.
    # NOTE: Λ is a diagonal matrix, represented as a vector
    #       of the diagonal entries. Thus, the matrix dot product
    #       is represented nicely by the NumPy broadcasting of
    #       the *scalar* multiplication. In particular,
    #       Q · Λ = Q_ij Λ_jk = Q_ij δ_jk λ_k = Q_ij λ_j.
    #       As NumPy arrays, Q has shape (N, N) and
    #       Lambda has shape (N, ), such that the broadcasting
    #       represents precisely Q_ij λ_j.
    return Q * Lambda


def rand_ket(dimensions, density=1, distribution="haar", *,
             seed=None, dtype=_data.Dense):
    """Creates a random ket vector.

    Parameters
    ----------
    dimensions : (int) or (list of int) or (list of list of int)
        Dimension of Hilbert space. If provided as a list of ints, then the
        dimension is the product over this list, but the ``dims`` property of
        the new Qobj are set to this list.  This can produce either `oper` or
        `super` depending on the passed `dimensions`.

    density : float, [1]
        Density between [0,1] of output ket state when using the ``fill``
        method.

    distribution : str {"haar", "fill"}
        Method used to obtain the kets.

        - haar : Haar random pure state obtained by applying a Haar random
          unitary to a fixed pure state.
        - fill : Fill the ket with uniformly distributed random complex number.

    seed : int, SeedSequence, Generator, optional
        Seed to create the random number generator or a pre prepared
        generator. When none is suplied, a default generator is used.

    dtype : type or str
        Storage representation. Any data-layer known to `qutip.data.to` is
        accepted.

    Returns
    -------
    oper : qobj
        Ket quantum state vector.
    """
    generator = _get_generator(seed)
    N, dims = _implicit_tensor_dimensions(dimensions)
    if distribution not in ["haar", "fill"]:
        raise ValueError("distribution must be one of {'haar', 'fill'}")

    if distribution == "haar":
        ket = rand_unitary(N, density, "haar", seed=generator) @ basis(N, 0)
    else:
        X = scipy.sparse.rand(N, 1, density, format='csr',
                              random_state=generator)
        while X.nnz == 0:
            # ensure that the ket is not all zeros.
            X = scipy.sparse.rand(N, 1, density+1/N, format='csr',
                                  random_state=generator)
        X.data = X.data - 0.5
        Y = X.copy()
        Y.data = 1.0j * (generator.random(len(X.data)) - 0.5)
        X = _data.csr.CSR(X + Y)
        ket = Qobj(_data.mul(X, 1 / _data.norm.l2(X)),
                   copy=False, type="ket", isherm=False, isunitary=False)
    ket.dims = [dims[0], [1]]
    return ket.to(dtype)


def rand_dm(dimensions, density=0.75, distribution="ginibre", *,
            eigenvalues=(), rank=None, seed=None, dtype=_data.CSR):
    r"""Creates a random density matrix of the desired dimensions.

    Parameters
    ----------
    dimensions : (int) or (list of int) or (list of list of int)
        Dimension of Hilbert space. If provided as a list of ints, then the
        dimension is the product over this list, but the ``dims`` property of
        the new Qobj are set to this list.  This can produce either ``oper`` or
        ``super`` depending on the passed ``dimensions``.

    density : float
        Density between [0,1] of output density matrix. Used by the "pure",
        "eigen" and "herm".

    distribution : str {"ginibre", "hs", "pure", "eigen", "uniform"}
        Method used to obtain the density matrices.

        - "ginibre" : Ginibre random density operator of rank ``rank`` by using
          the algorithm of [BCSZ08]_.
        - "hs" : Hilbert-Schmidt ensemble, equivalent to a full rank ginibre
          operator.
        - "pure" : Density matrix created from a random ket.
        - "eigen" : A density matrix with the given ``eigenvalues``.
        - "herm" : Build from a random hermitian matrix using ``rand_herm``.

    eigenvalues : array_like, optional
        Eigenvalues of the output Hermitian matrix. The len must match the
        shape of the matrix.

    rank : int, optional
        When using the "ginibre" distribution, rank of the density matrix.
        Will default to a full rank operator when not provided.

    seed : int, SeedSequence, Generator, optional
        Seed to create the random number generator or a pre prepared
        generator. When none is suplied, a default generator is used.

    dtype : type or str
        Storage representation. Any data-layer known to `qutip.data.to` is
        accepted.

    Returns
    -------
    oper : qobj
        Density matrix quantum operator.
    """
    generator = _get_generator(seed)
    N, dims = _implicit_tensor_dimensions(dimensions)
    distributions = set(["eigen", "ginibre", "hs", "pure", "herm"])
    if distribution not in distributions:
        raise ValueError(f"distribution must be one of {distributions}")

    if distribution == "eigen":
        if len(eigenvalues) != N:
            raise ValueError("Number of eigenvalues does not match the shape.")
        if np.abs(np.sum(eigenvalues)-1.0) > 1e-15 * N:
            raise ValueError('Eigenvalues of a density matrix '
                             f'must sum to one, not {np.sum(eigenvalues)}')
        H = _data.diag(eigenvalues, 0)
        nvals = N**2 * density
        H = _rand_jacobi_rotation(H, generator)
        while _data.csr.nnz(H) < 0.95*nvals:
            H = _rand_jacobi_rotation(H, generator)
    elif distribution == "ginibre":
        H = _rand_dm_ginibre(N, rank, generator)
    elif distribution == "hs":
        H = _rand_dm_ginibre(N, N, generator)
    elif distribution == "pure":
        dm_density = np.sqrt(density)
        psi = rand_ket(N, density=dm_density,
                       distribution="fill", seed=generator)
        H = psi.proj().data
    elif distribution == "herm":
        block_sizes = _get_block_sizes(N, density, generator)
        blocks = []
        first = True
        for block in block_sizes:
            if block != 1:
                mat = rand_herm(block, 1, seed=generator, dtype="dense").full()
            else:
                mat = np.ones((1,1))
                if not first:
                    mat *= max(generator.random() - 0.5, 0)
                first = False
            blocks.append(mat.T.conj() @ mat)
        H = _merge_shuffle_blocks(blocks, generator)
        H /= H.trace()

    return Qobj(H, dims=dims, type='oper', isherm=True, copy=False).to(dtype)


def _rand_dm_ginibre(N, rank, generator):
    """
    Returns a Ginibre random density operator of  rank ``rank`` by using the
    algorithm of [BCSZ08]_. If ``rank`` is `None`, a full-rank
    (Hilbert-Schmidt ensemble) random density operator will be
    returned.

    Parameters
    ----------
    N : int
        Dimension of the density operator to be returned.

    rank : int or None
        Rank of the sampled density operator. If None, a full-rank
        density operator is generated.

    Returns
    -------
    rho : Qobj
        An N × N density operator sampled from the Ginibre
        or Hilbert-Schmidt distribution.
    """
    if not rank:
        rank = N
    if rank > N:
        raise ValueError("Rank cannot exceed dimension.")

    X = _randnz((N, rank), generator, norm='ginibre')
    rho = np.dot(X, X.T.conj())
    rho /= np.trace(rho)

    return rho


def rand_kraus_map(dimensions, *, seed=None, dtype=_data.Dense):
    """
    Creates a random CPTP map on an N-dimensional Hilbert space in Kraus
    form.

    Parameters
    ----------
    dimensions : (int) or (list of int) or (list of list of int)
        Dimension of Hilbert space. If provided as a list of ints, then the
        dimension is the product over this list, but the ``dims`` property of
        the new Qobj are set to this list.  This can produce either `oper` or
        `super` depending on the passed `dimensions`.

    seed : int, SeedSequence, Generator, optional
        Seed to create the random number generator or a pre prepared
        generator. When none is suplied, a default generator is used.

    dtype : type or str
        Storage representation. Any data-layer known to `qutip.data.to` is
        accepted.

    Returns
    -------
    oper_list : list of qobj
        N^2 x N x N qobj operators.

    """
    N, dims = _implicit_tensor_dimensions(dimensions)

    # Random unitary (Stinespring Dilation)
    big_unitary = rand_unitary(N ** 3, seed=seed, dtype=dtype).full()
    orthog_cols = np.array(big_unitary[:, :N])
    oper_list = np.reshape(orthog_cols, (N ** 2, N, N))
    return [Qobj(x, dims=dims, type='oper', copy=False).to(dtype)
            for x in oper_list]


def rand_super(dimensions, *, superrep="super", seed=None, dtype=_data.Dense):
    """
    Returns a randomly drawn superoperator acting on operators acting on
    N dimensions.

    Parameters
    ----------
    dimensions : (int) or (list of int) or (list of list of int)
        Dimension of Hilbert space. If provided as a list of ints, then the
        dimension is the product over this list, but the ``dims`` property of
        the new Qobj are set to this list.  This can produce either `oper` or
        `super` depending on the passed `dimensions`.

    superrop : str, optional, {"super"}
        Representation of the super operator

    seed : int, SeedSequence, Generator, optional
        Seed to create the random number generator or a pre prepared
        generator. When none is suplied, a default generator is used.

    dtype : type or str
        Storage representation. Any data-layer known to `qutip.data.to` is
        accepted.
    """
    generator = _get_generator(seed)
    from .solver.propagator import propagator
    N, dims = _implicit_tensor_dimensions(dimensions, superoper=True)
    H = rand_herm(N, seed=generator, dtype=dtype)
    S = propagator(H, generator.random(), [
        create(N), destroy(N), jmat(float(N - 1) / 2.0, 'z')
    ])
    S.dims = dims
    out = {
            "choi" : to_choi,
            "chi" : to_chi,
            "super": to_super,
        }[superrep](S).to(dtype)
    return out


def rand_super_bcsz(dimensions, enforce_tp=True, rank=None, *,
                    superrep="super", seed=None, dtype=_data.CSR):
    """
    Returns a random superoperator drawn from the Bruzda
    et al ensemble for CPTP maps [BCSZ08]_. Note that due to
    finite numerical precision, for ranks less than full-rank,
    zero eigenvalues may become slightly negative, such that the
    returned operator is not actually completely positive.

    Parameters
    ----------
    dimensions : (int) or (list of int) or (list of list of int)
        Dimension of Hilbert space. If an int is provided, it is understood as
        the Square root of the dimension of the superoperator to be returned,
        with the corresponding dims as ``[[[N],[N]], [[N],[N]]]``. If provided
        as a list of ints, then the dimensions is understood as the space of
        density matrices this superoperator is applied to: ``dimensions=[2,2]``
        ``dims=[[[2,2],[2,2]], [[2,2],[2,2]]]``.

    enforce_tp : bool
        If True, the trace-preserving condition of [BCSZ08]_ is enforced;
        otherwise only complete positivity is enforced.

    rank : int or None
        Rank of the sampled superoperator. If None, a full-rank
        superoperator is generated.

    seed : int, SeedSequence, Generator, optional
        Seed to create the random number generator or a pre prepared
        generator. When none is suplied, a default generator is used.

    superrop : str, optional, {"super"}
        representation of the

    dtype : type or str
        Storage representation. Any data-layer known to `qutip.data.to` is
        accepted.

    Returns
    -------
    rho : Qobj
        A superoperator acting on vectorized dim × dim density operators,
        sampled from the BCSZ distribution.
    """
    generator = _get_generator(seed)
    N, dims = _implicit_tensor_dimensions(dimensions, superoper=True)

    if rank is None:
        rank = N**2
    if rank > N**2:
        raise ValueError("Rank cannot exceed superoperator dimension.")

    # We use mainly dense matrices here for speed in low
    # dimensions. In the future, it would likely be better to switch off
    # between sparse and dense matrices as the dimension grows.

    # We start with a Ginibre uniform matrix X of the appropriate rank,
    # and use it to construct a positive semidefinite matrix X X⁺.
    X = _randnz((N**2, rank), generator, norm='ginibre')

    # Precompute X X⁺, as we'll need it in two different places.
    XXdag = np.dot(X, X.T.conj())
    tmp_dims = [[[N], [N]], [[N], [N]]]

    if enforce_tp:
        # We do the partial trace over the first index by using dense reshape
        # operations, so that we can avoid bouncing to a sparse representation
        # and back.
        Y = np.einsum('ijik->jk', XXdag.reshape((N, N, N, N)))

        # Now we have the matrix 𝟙 ⊗ Y^{-1/2}, which we can find by doing
        # the square root and the inverse separately. As a possible
        # improvement, iterative methods exist to find inverse square root
        # matrices directly, as this is important in statistics.
        Z = np.kron(
            np.eye(N),
            scipy.linalg.sqrtm(scipy.linalg.inv(Y))
        )

        # Finally, we dot everything together and pack it into a Qobj,
        # marking the dimensions as that of a type=super (that is,
        # with left and right compound indices, each representing
        # left and right indices on the underlying Hilbert space).
        D = Qobj(np.dot(Z, np.dot(XXdag, Z)), dims=tmp_dims, type='super')
    else:
        D = N * Qobj(XXdag / np.trace(XXdag), dims=tmp_dims, type='super')

    # Since [BCSZ08] gives a row-stacking Choi matrix, but QuTiP
    # expects a column-stacking Choi matrix, we must permute the indices.
    D = D.permute([[1], [0]])

    # Mark that we've made a Choi matrix.
    D.superrep = 'choi'
    D.dims = dims
    out = {
            "choi" : to_choi,
            "chi" : to_chi,
            "super": to_super,
        }[superrep](D).to(dtype)
    return out


def rand_stochastic(dimensions, density=0.75, kind='left',
                    *, seed=None, dtype=_data.CSR):
    """Generates a random stochastic matrix.

    Parameters
    ----------
    dimensions : (int) or (list of int) or (list of list of int)
        Dimension of Hilbert space. If provided as a list of ints, then the
        dimension is the product over this list, but the ``dims`` property of
        the new Qobj are set to this list.  This can produce either `oper` or
        `super` depending on the passed `dimensions`.

    density : float, [0.75]
        Density between [0,1] of output density matrix.

    kind : str (Default = 'left')
        Generate 'left' or 'right' stochastic matrix.

    seed : int, SeedSequence, Generator, optional
        Seed to create the random number generator or a pre prepared
        generator. When none is suplied, a default generator is used.

    dtype : type or str
        Storage representation. Any data-layer known to `qutip.data.to` is
        accepted.

    Returns
    -------
    oper : qobj
        Quantum operator form of stochastic matrix.
    """
    generator = _get_generator(seed)
    N, dims = _implicit_tensor_dimensions(dimensions)
    num_elems = max([int(np.ceil(N*(N+1)*density)/2), N])
    data = generator.random(num_elems)
    # Ensure an element on every row and column
    row_idx = np.hstack([generator.permutation(N),
                         generator.choice(N, num_elems-N)])
    col_idx = np.hstack([generator.permutation(N),
                         generator.choice(N, num_elems-N)])
    M = sp.coo_matrix((data, (row_idx, col_idx)),
                      dtype=np.complex128, shape=(N, N)).tocsr()
    M = 0.5 * (M + M.conj().transpose())
    num_rows = M.indptr.shape[0] - 1
    for row in range(num_rows):
        row_start = M.indptr[row]
        row_end = M.indptr[row+1]
        row_sum = np.sum(M.data[row_start:row_end])
        M.data[row_start:row_end] /= row_sum
    if kind == 'left':
        M = M.transpose()
    return Qobj(M, dims=dims).to(dtype)
